from zs_utils.api.etsy.base_api import EtsyAPI


class GetEtsyListingImageList(EtsyAPI):
    """
    Docs:
    https://www.etsy.com/openapi/developers#operation/getListingImages
    """

    http_method = "GET"
    resource_method = "shops/{shop_id}/listings/{listing_id}/images"
    required_params = ["shop_id", "listing_id"]


class GetEtsyListingImage(EtsyAPI):
    """
    Docs:
    https://www.etsy.com/openapi/developers#operation/getListingImage
    """

    http_method = "GET"
    resource_method = "shops/{shop_id}/listings/{listing_id}/images/{listing_image_id}"
    required_params = ["shop_id", "listing_id", "listing_image_id"]


class DeleteEtsyListingImage(EtsyAPI):
    """
    Docs:
    https://www.etsy.com/openapi/developers#operation/deleteListingImage
    """

    http_method = "DELETE"
    resource_method = "shops/{shop_id}/listings/{listing_id}/images/{listing_image_id}"
    required_params = ["shop_id", "listing_id", "listing_image_id"]


class UploadEtsyListingImage(EtsyAPI):
    """
    Docs:
    https://www.etsy.com/openapi/developers#operation/uploadListingImage
    """

    http_method = "POST"
    resource_method = "shops/{shop_id}/listings/{listing_id}/images"
    required_params = ["shop_id", "listing_id"]
    allowed_params = [
        "image",
        "listing_image_id",
        "rank",
        "overwrite",
        "is_watermarked",
    ]

    @property
    def headers(self) -> dict:
        # При наличии файла будет автоматом подставлен multipart - что нам и нужно
        headers = super().headers
        headers.pop("Content-Type")
        return headers


class GetVariationImages(EtsyAPI):
    """
    Docs:
    https://www.etsy.com/openapi/developers#operation/getListingVariationImages
    """

    http_method = "GET"
    resource_method = "shops/{shop_id}/listings/{listing_id}/variation-images"
    required_params = ["shop_id", "listing_id"]


class UpdateVariationImages(EtsyAPI):
    """
    Docs:
    https://www.etsy.com/openapi/developers#operation/updateVariationImages
    """

    http_method = "POST"
    resource_method = "shops/{shop_id}/listings/{listing_id}/variation-images"
    required_params = ["shop_id", "listing_id", "variation_images"]



from zs_utils.api.wildberries.base_api import WildberriesAPI


class GetWildberriesOrders(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1orders/get
    """

    http_method = "GET"
    resource_method = "api/v3/orders"
    required_params = [
        "limit",  # 1000
        "next",  # offset, 0 for the first request
    ]
    allowed_params = [
        "dateFrom",
        "dateTo",
    ]


class GetWildberriesNewOrders(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1orders~1new/get
    """

    http_method = "GET"
    resource_method = "api/v3/orders/new"


class CancelWildberriesOrder(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1orders~1%7BorderId%7D~1cancel/patch
    """

    http_method = "PATCH"
    resource_method = "api/v3/orders/{order_id}/cancel"
    required_params = [
        "order_id",
    ]


class GetWildberriesOrdersStatus(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1orders~1status/post
    """

    http_method = "POST"
    resource_method = "api/v3/orders/status"
    required_params = [
        "orders",
    ]


class DeleteWildberriesSupply(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1supplies~1%7BsupplyId%7D/delete
    """

    http_method = "DELETE"
    resource_method = "api/v3/supplies/{supply_id}"
    required_params = [
        "supply_id",
    ]


class GetWildberriesSupplies(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1supplies/get
    """

    http_method = "GET"
    resource_method = "api/v3/supplies"
    required_params = [
        "limit",  # Up to 1000
        "next",  # offset, 0 for the first request
    ]


class GetWildberriesSupplyOrders(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1supplies~1%7BsupplyId%7D~1orders/get
    """

    http_method = "GET"
    resource_method = "api/v3/supplies/{supply_id}/orders"
    required_params = [
        "supply_id",
    ]


class GetWildberriesSupplyBarcode(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1supplies~1%7BsupplyId%7D~1barcode/get
    """

    http_method = "GET"
    resource_method = "api/v3/supplies/{supply_id}/barcode"
    required_params = [
        "supply_id",
        "type",
    ]


class GetWildberriesOrdersStickers(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1orders~1stickers/post
    """

    http_method = "POST"
    resource_method = "api/v3/orders/stickers"
    required_params = [
        "type",
        "width",
        "height",
        "orders",
    ]


class CreateWildberriesSupply(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1supplies/post
    """

    http_method = "POST"
    resource_method = "api/v3/supplies"
    allowed_params = ["name"]


class AddOrdersToWildberriesSupply(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1supplies~1%7BsupplyId%7D~1orders~1%7BorderId%7D/patch
    """

    http_method = "PATCH"
    resource_method = "api/v3/supplies/{supply_id}/orders/{order_id}"
    required_params = [
        "supply_id",
        "order_id",
    ]


class CloseWildberriesSupply(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1supplies~1%7BsupplyId%7D~1deliver/patch
    """

    http_method = "PATCH"
    resource_method = "api/v3/supplies/{supply_id}/deliver"
    required_params = [
        "supply_id",
    ]


class AddWildberriesSgtinToOrder(WildberriesAPI):
    """
    https://openapi.wb.ru/#tag/Marketplace-Sborochnye-zadaniya/paths/~1api~1v3~1orders~1%7BorderId%7D~1meta~1sgtin/post
    """

    http_method = "POST"
    resource_method = "api/v3/orders/{order_id}/meta/sgtin"
    required_params = [
        "order_id",
        "sgtin",
    ]



#!/usr/bin/env python
# coding=utf-8

"""Converter object to handle string input."""

from decimal import Decimal as D

from zs_utils.unit_converter.parser import QuantityParser, UnitParser


def convertd(quantity: str, desired_unit: str) -> D:
    """

    :param quantity:
    :param desired_unit:
    :return:

    Examples :
    ----------

    >>> from unit_converter import convert
    >>> convert('2.78 daN*mm^2', 'mN*µm^2')
    Decimal('2.78E+10')
    """
    quantity = QuantityParser().parse(quantity)
    desired_unit = UnitParser().parse(desired_unit)
    return quantity.convert(desired_unit).value


def converts(quantity: str, desired_unit: str) -> str:
    """

    :param quantity:
    :param desired_unit:
    :return:

    Examples :
    ----------

    >>> from unit_converter import converts
    >>> converts('2.78 daN*mm^2', 'mN*µm^2')
    '2.78E+10'
    """
    return str(convertd(quantity=quantity, desired_unit=desired_unit))


def convert(value: str, from_unit: str, to_unit: str) -> float:
    """

    :param value:
    :param from_unit:
    :param to_unit:
    :return:

    Examples :
    ----------

    >>> from unit_converter import convert
    >>> convertf('2.78', 'daN*mm^2', 'mN*µm^2')
    27800000000.0
    """

    # Тривиальные случаи
    if (float(value) == 0) or (from_unit == to_unit):
        return float(value)

    return float(convertd(quantity=f"{value} {from_unit}", desired_unit=to_unit))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from yookassa import Refund

from .payment_get import RemoteGetYooKassaPayment
from zs_utils.api.yookassa.base_action import YooKassaAction
from zs_utils.api.yookassa import services


__all__ = [
    "RemoteGetYooKassaRefund",
    "RemoteGetYooKassaRefundList",
    "RemoteCreateYooKassaRefund",
]


class RemoteGetYooKassaRefund(YooKassaAction):
    description = "Получение возврата платежа в Яндекс.Кассе"
    api_method = Refund.find_one
    required_params = ["refund_id"]


class RemoteGetYooKassaRefundList(YooKassaAction):
    description = "Получение всех возвратов платежей в Яндекс.Кассе"
    api_method = Refund.list
    required_params = ["params"]


class RemoteCreateYooKassaRefund(YooKassaAction):
    description = "Создание возврата платежа в Яндекс.Кассе"
    USER_REQUIRED = True
    api_method = Refund.create
    required_params = ["params"]

    def set_used_actions(self) -> None:
        self.remote_get_payment = self.set_used_action(action_class=RemoteGetYooKassaPayment)

    def get_params(
        self,
        payment_id: str,
        amount: float = None,
        description: str = None,
        receipt: dict = None,
        extra_metadata: dict = None,
        **kwargs,
    ):
        # Получение информации из сервиса
        payment_data = self.remote_get_payment(payment_id=payment_id)[2]["results"]

        currency = payment_data["amount"]["currency"]
        if not description:
            description = payment_data["description"]
        if not amount:
            amount = payment_data["amount"]["value"]

        if not receipt:
            receipt = services.BaseYooKassaService.get_receipt_template(
                product_description=description,
                product_price=amount,
                product_price_currency=currency,
                user=self.user,
            )

        kwargs["params"] = {
            "amount": {"value": amount, "currency": currency},
            "payment_id": payment_id,
            "receipt": receipt,
            "metadata": extra_metadata,
        }

        return super().get_params(**kwargs)



from channels import layers
from asgiref.sync import async_to_sync

from django.conf import settings

from zs_utils.json_utils import custom_json_dumps
from zs_utils.websocket import consumer


__all__ = [
    "WebsocketService",
]


class WebsocketService:
    """
    Сервис для отправки сообщений в вебсокет.
    """

    @classmethod
    def send_data_to_consumer_group(cls, *, group_name: str, content: dict) -> None:
        """
        Отправить данные content в websocket
        """
        layer = layers.get_channel_layer()
        async_to_sync(layer.group_send)(
            group_name,
            {
                "type": "send.message",
                "content": custom_json_dumps(content, ensure_ascii=True),
            },
        )

    @classmethod
    def send_data_to_user(cls, *, user_id: int, content: dict) -> None:
        """
        Отправить сообщение пользователю.
        """

        group_name = consumer.BaseChannelsConsumer.get_user_group_name(user_id=user_id)
        cls.send_data_to_consumer_group(group_name=group_name, content=content)

    @classmethod
    def send_page_refresh_required_message(cls, user_id: int, page_path: str) -> None:
        """
        Отправление пользователю сообщения о необходимости обновить страницу.
        """

        group_name = consumer.BaseChannelsConsumer.get_user_group_name(user_id=user_id)
        cls.send_data_to_consumer_group(
            group_name=group_name,
            content={"event_type": "page_refresh_required", "page": page_path},
        )

    @classmethod
    def send_data_to_notification_group(cls, *, user_id: int, notification: str, data: dict) -> None:
        """
        Отправить сообщение пользователю, если он подписан на группу событий.
        """

        if notification not in getattr(settings, "WEBSOCKET_NOTIFICATION_GROUPS", []):
            raise ValueError("Недопустимая группа событий.")

        group_name = consumer.BaseChannelsConsumer.get_user_notification_group_name(
            user_id=user_id,
            notification=notification,
        )
        cls.send_data_to_consumer_group(group_name=group_name, content={"notification": notification, "data": data})



from zs_utils.api.fedex.base_api import FedexAPI


class FedexCreateShipment(FedexAPI):
    http_method = "POST"
    resource_method = "/ship/v1/shipments"
    required_params = [
        "requestedShipment",
        "labelResponseOptions",
        "accountNumber",
    ]
    allowed_params = [
        "mergeLabelDocOption",
        "shipAction",
        "processingOptionType",
        "oneLabelAtATime",
    ]


class FedexCancelShipment(FedexAPI):
    http_method = "PUT"
    resource_method = "/ship/v1/shipments/cancel"
    required_params = [
        "accountNumber",
        "trackingNumber",
    ]
    allowed_params = [
        "senderCountryCode",
        "deletionControl",
    ]



from .aspect_enums import *
from .common_enums import *
from .listing_enums import *
from .message_enums import *
from .notification_enums import *
from .order_enums import *
from .policy_enums import *
from .shipping_carriers_enum import *



from zs_utils.api.insales.base_api import InsalesAPI


class CreateInsalesOrder(InsalesAPI):
    """
    https://api.insales.ru/#order-create-order-with-product-json
    """

    resource_method = "orders.json"
    http_method = "POST"
    required_params = ["order"]


class DeleteInsalesOrder(InsalesAPI):
    """
    В доках нет, но догадаться как это сделать было нетрудно
    """

    resource_method = "orders/{order_id}.json"
    http_method = "DELETE"
    required_params = ["order_id"]



from zs_utils.api.etsy.base_api import EtsyAPI


class GetEtsyShopSectionList(EtsyAPI):
    """
    https://www.etsy.com/openapi/developers#operation/getShopSections
    """

    http_method = "GET"
    resource_method = "shops/{shop_id}/sections"
    required_params = ["shop_id"]


class GetEtsyShopSection(EtsyAPI):
    """
    https://www.etsy.com/openapi/developers#operation/getShopSection
    """

    http_method = "GET"
    resource_method = "shops/{shop_id}/sections/{shop_section_id}"
    required_params = ["shop_id", "shop_section_id"]


class CreateEtsyShopSection(EtsyAPI):
    """
    https://www.etsy.com/openapi/developers#operation/createShopSection
    """

    http_method = "POST"
    resource_method = "shops/{shop_id}/sections"
    required_params = ["shop_id", "title"]



from zs_utils.api.ebay.base_api import EbayAPI


class CreatePaymentPolicy(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/payment_policy/methods/createPaymentPolicy
    """

    http_method = "POST"
    resource_method = "sell/account/v1/payment_policy"


class DeletePaymentPolicy(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/payment_policy/methods/deletePaymentPolicy
    """

    http_method = "DELETE"
    resource_method = "sell/account/v1/payment_policy/{payment_policy_id}"
    required_params = ["payment_policy_id"]


class GetEbayPaymentPolicy(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/payment_policy/methods/getPaymentPolicy
    """

    http_method = "GET"
    resource_method = "sell/account/v1/payment_policy/{payment_policy_id}"
    required_params = ["payment_policy_id"]


class GetPaymentPolicies(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/payment_policy/methods/getPaymentPolicies
    """

    http_method = "GET"
    resource_method = "sell/account/v1/payment_policy"
    required_params = ["marketplace_id"]


class GetPaymentPolicyByName(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/payment_policy/methods/getPaymentPolicyByName
    """

    http_method = "GET"
    resource_method = "sell/account/v1/payment_policy/get_by_policy_name"
    required_params = ["marketplace_id", "name"]


class UpdateEbayPaymentPolicy(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/payment_policy/methods/updatePaymentPolicy
    """

    http_method = "PUT"
    resource_method = "sell/account/v1/payment_policy/{payment_policy_id}"
    required_params = ["payment_policy_id"]



from zs_utils.api.base_api import BaseAPI


class EtsyAPI(BaseAPI):
    production_api_url = "https://openapi.etsy.com/v3/application/"

    def __init__(self, access_token: str, api_key: str, **kwargs):
        super().__init__(**kwargs)

        self.access_token = access_token
        self.api_key = api_key

    @property
    def headers(self) -> dict:
        headers = {
            "Authorization": f"Bearer {self.access_token}",
            "x-api-key": self.api_key,
        }
        if self.http_method in ["POST", "PUT", "PATCH"]:
            headers.update(
                {
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                    "Accept-Charset": "utf-8",
                    "Accept-Encoding": "application/gzip",
                }
            )
        return headers

    @staticmethod
    def get_user_id(access_token: str):
        return int(access_token.split(".")[0])

    def build_url(self, params: dict) -> str:
        if "user_id" in self.path_params:
            params["user_id"] = self.get_user_id(access_token=self.access_token)
        return super().build_url(params)



from zs_utils.api.apiship.base_api import ApishipAPI


class ApishipGetPointsAPI(ApishipAPI):
    """
    Docs: https://docs.apiship.ru/docs/api/lists-points/
    """

    resource_method = "lists/points"
    http_method = "GET"
    allowed_params = [
        "limit",
        "offset",
        "filter",
        "fields",
        "stateCheckOff",
    ]



import base64
import simplejson
import logging
import urllib
import requests
from zs_utils.exceptions import CustomException

from django.utils.translation import gettext as _

from .model import environment, oAuth_token


class EbayOAuthClientError(CustomException):
    pass


class EbayOAuthClient:
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        redirect_uri: str,
        is_sandbox: bool,
        user_scopes: list = None,
        app_scopes: list = None,
        logger=None,
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri

        if is_sandbox:
            self.env_type = environment.SANDBOX
        else:
            self.env_type = environment.PRODUCTION

        self.user_scopes = user_scopes
        self.app_scopes = app_scopes

        self.headers = self._generate_request_headers()
        self.logger = logger or logging.getLogger(__name__)

    def _generate_request_headers(self) -> dict:
        b64_string = f"{self.client_id}:{self.client_secret}".encode()
        b64_encoded_credential = base64.b64encode(b64_string).decode("utf-8")
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": f"Basic {b64_encoded_credential}",
        }
        return headers

    def generate_user_authorization_url(self, state: str = None) -> str:
        param = {
            "client_id": self.client_id,
            "redirect_uri": self.redirect_uri,
            "response_type": "code",
            "prompt": "login",
            "scope": " ".join(self.user_scopes),
        }

        if state is not None:
            param.update({"state": state})

        query = urllib.parse.urlencode(param)
        return f"{self.env_type.web_endpoint}?{query}"

    def _get_token(self, data: dict) -> oAuth_token:
        response = requests.post(self.env_type.api_endpoint, data=data, headers=self.headers)
        content = simplejson.loads(response.content)

        status = response.status_code
        if status != requests.codes.ok:
            raise EbayOAuthClientError(
                message=_("Не удалось получить токен: {error}").format(error=content["error_description"])
            )

        return oAuth_token(**content)

    def get_application_access_token(self) -> oAuth_token:
        body = {
            "grant_type": "client_credentials",
            "redirect_uri": self.redirect_uri,
            "scope": " ".join(self.app_scopes),
        }
        return self._get_token(data=body)

    def exchange_code_for_access_token(self, code: str) -> oAuth_token:
        body = {
            "grant_type": "authorization_code",
            "redirect_uri": self.redirect_uri,
            "code": code,
        }
        return self._get_token(data=body)

    def get_user_access_token(self, refresh_token: str) -> oAuth_token:
        body = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "scope": " ".join(self.user_scopes),
        }
        return self._get_token(data=body)



from zs_utils.api.aliexpress_russia.base_api import AliexpressRussiaAPI


class GetAliexpressListingListAPI(AliexpressRussiaAPI):
    """
    https://business.aliexpress.ru/docs/product-list#heading-poluchit-otfiltrovannii-spisok-tovarov
    """

    resource_method = "/api/v1/scroll-short-product-by-filter"
    required_params = ["limit"]  # 50
    allowed_params = [
        "filter",
        "last_product_id",
    ]


class GetAliexpressListingAPI(AliexpressRussiaAPI):
    """
    https://business.aliexpress.ru/docs/product-list#heading-poluchit-informatsiyu-o-tovare
    """

    resource_method = "/api/v1/product/get-seller-product"
    required_params = ["product_id"]


class CreateAliexpressListingAPI(AliexpressRussiaAPI):
    """
    https://business.aliexpress.ru/docs/local-create-products
    """

    resource_method = "/api/v1/product/create"
    required_params = ["products"]


class UpdateAliexpressListingAPI(AliexpressRussiaAPI):
    """
    https://business.aliexpress.ru/docs/post-products#heading-obnovit-vse-polya-tovara
    """

    resource_method = "/api/v1/product/edit"
    required_params = ["products"]


class SetPublishedAliexpressListingAPI(AliexpressRussiaAPI):
    """
    https://business.aliexpress.ru/docs/post-products#heading-vernut-tovari-v-prodazhu
    """

    resource_method = "/api/v1/product/online"
    required_params = ["productIds"]  # 20


class SetUnpublishedAliexpressListingAPI(AliexpressRussiaAPI):
    """
    https://business.aliexpress.ru/docs/post-products#heading-snyat-tovari-s-prodazhi
    """

    resource_method = "/api/v1/product/offline"
    required_params = ["productIds"]  # 20


class GetAliexpressListingStatusAPI(AliexpressRussiaAPI):
    """
    https://business.aliexpress.ru/docs/post-products#heading-poluchit-statusi-zagruzkiobnovleniya-tovara
    """

    http_method = "GET"
    resource_method = "/api/v1/tasks"
    required_params = ["group_id"]


class GetAliexpressProductStocksAPI(AliexpressRussiaAPI):
    """
    https://business.aliexpress.ru/docs/manage-stocks#heading-poluchit-ostatki-tovara
    """

    resource_method = "/api/v1/stocks"
    required_params = ["stocks"]


class UpdateAliexpressProductStocksAPI(AliexpressRussiaAPI):
    """
    https://business.aliexpress.ru/docs/manage-stocks#heading-obnovit-ostatki-tovara-po-sku-kodu
    """

    resource_method = "/api/v1/product/update-sku-stock"
    required_params = ["products"]


class UpdateAliexpressProductPriceAPI(AliexpressRussiaAPI):
    """
    https://business.aliexpress.ru/docs/manage-stocks#heading-obnovit-tsenu-tovara-po-sku-kodu
    """

    resource_method = "/api/v1/product/update-sku-price"
    required_params = ["products"]


class GetAliexpressShippingTemplatesAPI(AliexpressRussiaAPI):
    """
    https://business.aliexpress.ru/docs/local-get-shipping-templates
    """

    http_method = "GET"
    resource_method = "/api/v1/sellercenter/get-count-product-on-onboarding-template"



from model_utils import Choices

from django.utils.translation import gettext_lazy as _


HTTP_METHODS = Choices("GET", "POST", "DELETE", "PUT", "PATCH")

API_ERROR_REASONS = Choices(
    ("system", _("Системная ошибка")),
    ("invalid_token", _("Проблемы с валидацией токена")),
    ("data_validation", _("Проблема с валидацией данных")),
    ("unknown", _("Неизвестная ошибка")),
    ("account_deactivated", _("Аккаунт отключен")),
    ("request_limit", _("Превышен лимит запросов")),
    ("object_not_found", _("Запрашиваемый объект не найден")),
    ("object_already_exists", _("Объект уже существует")),
    ("timeout", _("Время ожидания ответа на запрос истекло")),
    ("account_token_expired", _("Требуется обновление токена аккаунта")),
)



from zs_utils.api.shopify.base_api import ShopifyAPI


class GetShopifyFulfillmentAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/fulfillment#get-orders-order-id-fulfillments-fulfillment-id
    """

    http_method = "GET"
    resource_method = "orders/{order_id}/fulfillments/{fulfillment_id}.json"
    required_params = ["order_id", "fulfillment_id"]


class GetShopifyFulfillmentListAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/fulfillment#get-orders-order-id-fulfillments
    """

    http_method = "GET"
    resource_method = "orders/{order_id}/fulfillments.json"
    required_params = ["order_id"]
    allowed_params = ["page_info"]


class GetShopifyFulfillmentOrderListAPI(ShopifyAPI):
    """
    https://shopify.dev/docs/api/admin-rest/2023-10/resources/fulfillmentorder#get-orders-order-id-fulfillment-orders
    """

    http_method = "GET"
    resource_method = "orders/{order_id}/fulfillment_orders.json"
    required_params = ["order_id"]


class UpdateShopifyFulfillmentAPI(ShopifyAPI):
    """
    https://shopify.dev/docs/api/admin-rest/2023-10/resources/fulfillment#post-fulfillments-fulfillment-id-update-tracking
    """

    http_method = "PUT"
    resource_method = "fulfillments/{fulfillment_id}/update_tracking.json"
    payload_key = "fulfillment"
    required_params = ["fulfillment_id", "tracking_number"]


class CreateShopifyFulfillmentAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/fulfillment#post-orders-order-id-fulfillments
    """

    http_method = "POST"
    resource_method = "fulfillments.json"
    payload_key = "fulfillment"
    required_params = [
        "line_items_by_fulfillment_order",
    ]
    allowed_params = [
        "message",
        "notify_customer",
        "origin_address",
        "tracking_info",
    ]



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional, Union
from xml.etree import ElementTree as ET

from zs_utils.yandex_market_language.exceptions import ValidationError


XMLElement = ET.Element
XMLSubElement = ET.SubElement


class AbstractModel(ABC):
    """
    Abstract model for creating child models.
    """

    @abstractmethod
    def create_dict(self, **kwargs) -> dict:
        """
        Must be inherited by each child class.
        Describes the logic for creating a dictionary with data from a model.
        """
        raise NotImplementedError

    @abstractmethod
    def create_xml(self, **kwargs) -> XMLElement:
        """
        Must be inherited by each child class.
        Describes the logic for creating a xml with data from a model.
        """
        raise NotImplementedError

    @staticmethod
    def from_xml(el: XMLElement) -> "AbstractModel":
        """
        Must be inherited by each child class.
        Describes the logic for creating a model from an xml element.
        """
        raise NotImplementedError

    def to_xml(self, root_el: XMLElement = None) -> XMLElement:
        """
        Calls the inherited method to create the element and appends it to the
        parent, if it was set.
        """
        el = self.create_xml()
        if root_el is not None:
            root_el.append(el)
        return el

    def to_dict(self, clean: bool = False) -> dict:
        """
        Calls the inherited method to create the dictionary and returns a
        clean dictionary if the parameter was set.
        """
        return self.clean_dict if clean else self.create_dict()

    @property
    def clean_dict(self) -> dict:
        """
        A helper property to get clean dictionary with data.
        """
        return dict(**{k: v for k, v in self.create_dict().items() if v})

    @staticmethod
    def _is_valid_int(
        value, attr: str, allow_none: bool = False, convert_to_str: bool = True
    ) -> Optional[Union[int, str]]:
        """
        A helper method for checking if a value is a valid number and returning
        a value if the check succeeds or raising an error.
        """
        try:
            value = int(value)
            return str(value) if convert_to_str else value
        except (TypeError, ValueError):
            if value is None and allow_none:
                return None
            raise ValidationError("{a} must be a valid int".format(a=attr))

    @staticmethod
    def _is_valid_bool(value, attr: str, allow_none: bool = False) -> Optional[str]:
        """
        A helper method for checking if a value is a valid bool and returning
        a value if the check succeeds or raising an error.
        """
        if value in ["true", "false"]:
            return value
        elif value is True:
            return "true"
        elif value is False:
            return "false"
        elif value is None and allow_none:
            return None
        else:
            raise ValidationError(
                "The {attr} parameter should be boolean. " "Got {t} instead.".format(attr=attr, t=type(value))
            )

    @staticmethod
    def _is_valid_float(
        value, attr: str, allow_none: bool = False, convert_to_str: bool = True
    ) -> Optional[Union[float, str]]:
        """
        A helper method for checking if a value is a valid float and returning
        a value if the check succeeds or raising an error.
        """
        try:
            float(value)
            return str(value) if convert_to_str else value
        except (TypeError, ValueError):
            if value is None and allow_none:
                return None
            raise ValidationError("{a} must be a valid float".format(a=attr))

    @staticmethod
    def _str_to_bool(value: str) -> Optional[bool]:
        """
        Returns a boolean converted from a string or None.
        """
        if value == "true":
            return True
        elif value == "false":
            return False
        else:
            return None

    @staticmethod
    def _is_valid_datetime(
        dt,
        dt_format,
        attr: str,
        allow_none: bool = False,
    ) -> Optional[Union[datetime, str]]:
        """
        A helper method for checking if a value is a valid datetime and
        returning a value if the check succeeds or raising an error.
        """
        if isinstance(dt, datetime):
            return dt.strftime(dt_format)
        elif isinstance(dt, str):
            try:
                datetime.strptime(dt, dt_format)
            except ValueError:
                return None
                # raise ValidationError(e)
            return dt
        elif dt is None and allow_none:
            return None
        # else:
        #     raise ValidationError("{a} must be a valid datetime".format(a=attr))



from zs_utils.api.fedex.base_api import FedexAPI


class FedexTrackByTrackingNumber(FedexAPI):
    http_method = "POST"
    resource_method = "/track/v1/trackingnumbers"
    required_params = [
        "includeDetailedScans",
        "trackingInfo",
    ]



from yookassa import Payment

from zs_utils.api.yookassa.base_action import YooKassaAction


__all__ = [
    "RemoteGetYooKassaPayment",
    "RemoteGetYooKassaPaymentList",
]


class RemoteGetYooKassaPayment(YooKassaAction):
    description = "Получение платежа в Яндекс.Кассе"
    api_method = Payment.find_one
    required_params = ["payment_id"]


class RemoteGetYooKassaPaymentList(YooKassaAction):
    description = "Получение списка платежей в Яндекс.Кассе"
    api_method = Payment.list
    required_params = ["params"]



"""
Docs: https://developer.ebay.com/Devzone/XML/docs/Reference/eBay/types/NotificationEventTypeCodeType.html
"""
from model_utils import Choices


ClientAlertsEventTypeEnum = Choices(  # not compatible with PlatformNotifications
    "AccountSummary",
    "AccountSuspended",
    "EmailAddressChanged",
    "PasswordChanged",
    "PaymentDetailChanged",
)


PlatformNotificationEventTypeEnum = Choices(  # not compatible with ClientAlerts
    # BestOffers
    "BestOffer",
    # "BestOfferPlaced",
    # Feedbacks
    # "Feedback",
    # "FeedbackLeft",
    "FeedbackReceived",
    # Listings
    "FixedPriceTransaction",
    "ItemClosed",
    "ItemListed",
    "ItemOutOfStock",
    "ItemRevised",
    "ItemSuspended",
    # "ItemSold",
    # Messages
    "AskSellerQuestion",
    "M2MMessageStatusChange",
    "MyMessageseBayMessage",
    "MyMessageseBayMessageHeader",
    "MyMessagesHighPriorityMessage",
    "MyMessagesM2MMessage",
    "MyMessagesM2MMessageHeader",
    # Orders
    "ItemMarkedPaid",
    "ItemMarkedShipped",
    "BuyerResponseDispute",
    "BuyerCancelRequested",
    "OrderInquiryClosed",
    "OrderInquiryEscalatedToCase",
    "OrderInquiryOpened",
    "OrderInquiryProvideShipmentInformation",
    "OrderInquiryReminderForEscalation",
    "ReturnWaitingForSellerInfo",
    # Account
    "TokenRevocation",
    "UserIDChanged",
)


SupportedPlatformNotificationEventTypeEnum = Choices(
    "BestOffer",
    "FeedbackReceived",
    "ItemClosed",
    "ItemSuspended",
    "MyMessageseBayMessage",
    "MyMessagesM2MMessage",
    "UserIDChanged",
)



from .account import *
from .listing import *
from .carrier import *
from .category import *
from .receipt import *
from .shipping_profile_destination import *
from .shipping_profile_upgrade import *
from .shipping_profile import *
from .transaction import *



from zs_utils.api.insales.base_api import InsalesAPI


class GetInsalesProductList(InsalesAPI):
    """
    https://api.insales.ru/#product-get-products-json
    """

    resource_method = "products.json"
    http_method = "GET"
    allowed_params = [
        "updated_since",
        "from_id",
        "per_page",
        "with_deleted",
        "deleted",
    ]


class GetInsalesProductsCount(InsalesAPI):
    """
    https://api.insales.ru/#product-get-products-count-json
    """

    resource_method = "products/count.json"
    http_method = "GET"



from zs_utils.api.etsy.base_api import EtsyAPI


class FindAllShopTransactions(EtsyAPI):
    """
    Retrieves a set of Transaction objects associated to a Shop.

    Docs:
    https://www.etsy.com/openapi/developers#operation/getShopReceiptTransactionsByShop
    """

    http_method = "GET"
    api_target = "transactions"


class FindAllShopReceiptTransactions(EtsyAPI):
    """
    Retrieves a set of Transaction objects associated to a Shop_Receipt2.

    Docs:
    https://www.etsy.com/openapi/developers#operation/getShopReceiptTransactionsByReceipt
    """

    http_method = "GET"
    resource_method = "shops/{shop_id}/receipts/{receipt_id}/transactions"
    required_params = ["shop_id", "receipt_id"]



import requests

from zs_utils.api.constants import API_ERROR_REASONS
from zs_utils.api.base_action import APIAction
from zs_utils.api.amocrm import services


class AmocrmAction(APIAction):
    description = "Взаимодествие с API amoCRM"
    VALIDATE_PARAMS = True
    CUSTOM_FIELDS_MAPPING = None

    def __init__(self, app_id: str = None, **kwargs):
        super().__init__(**kwargs)

        # ID приложения
        self.app_id = app_id
        if app_id:
            self.app = services.AmocrmService.get_amocrm_app(app_id=app_id)
        else:
            self.app = services.AmocrmService.get_default_amocrm_app()

    def get_api_class_init_params(self, **kwargs) -> dict:
        if (not self.app.access_token) or self.app.access_token_expired:
            if (not self.app.refresh_token) or self.app.refresh_token_expired:
                raise self.exception_class("Токен обновления приложения amoCRM недействителен.")
            access_token = services.AmocrmService.refresh_amocrm_access_token(app_id=self.app_id)
        else:
            access_token = self.app.access_token
        return {"access_token": access_token}

    def get_params(self, **kwargs) -> dict:
        if self.CUSTOM_FIELDS_MAPPING:
            kwargs = self.process_custom_fields(data=kwargs, fields_mapping=self.CUSTOM_FIELDS_MAPPING)

        return super().get_params(**kwargs)

    def clean_api_request_params(self, raw_params: dict) -> dict:
        params = super().clean_api_request_params(raw_params)

        if getattr(self, "api_class", None) and self.api_class.array_payload:
            params = {"array_payload": [params]}

        return params

    def get_error_message(self, results: dict, response: requests.Response) -> str:
        message = ""

        if results.get("detail"):
            message = results["detail"]
        elif results.get("title"):
            message = results["title"]

        if results.get("hint"):
            message += ". " + results["hint"]

        if results.get("validation-errors"):
            for item in results["validation-errors"]:
                if item.get("errors"):
                    for error in item["errors"]:
                        message += error["path"] + ": " + error["detail"] + ". "

        return message

    def get_error_reason(self, results: dict, response: requests.Response, error_message: str) -> API_ERROR_REASONS:
        if results and results.get("hint") and (results["hint"] == "Token has expired"):
            return API_ERROR_REASONS.invalid_token
        return super().get_error_reason(results, response, error_message)

    @staticmethod
    def process_custom_fields(data: dict, fields_mapping: dict):
        result = {"custom_fields_values": []}

        for key, value in data.items():
            if key in fields_mapping:
                if isinstance(value, list):
                    if not value:
                        continue
                    values = [{"value": item} for item in value]
                else:
                    if (value is None) or (value == ""):
                        continue
                    values = [{"value": value}]
                result["custom_fields_values"].append({"field_id": fields_mapping[key], "values": values})
            else:
                result[key] = value

        if not result["custom_fields_values"]:
            result.pop("custom_fields_values")

        return result



import json
from datetime import datetime
from hashlib import md5
from zs_utils.api.base_api import BaseAPI


class AliexpressAPI(BaseAPI):
    production_api_url = "http://gw.api.taobao.com/router/rest"
    sandbox_url = "http://gw.api.tbsandbox.com/router/rest"
    http_method = "POST"

    def __init__(self, app_key: str, app_secret: str, access_token: str, **kwargs):
        super().__init__(**kwargs)

        self.app_key = app_key
        self.app_secret = app_secret
        self.access_token = access_token

    def build_url(self, params: dict) -> str:
        return self.sandbox_api_url if self.is_sandbox else self.production_api_url

    @property
    def headers(self) -> dict:
        return {
            "Content-Type": "application/x-www-form-urlencoded",
        }

    def add_sign(self, params):
        params.update(
            {
                "session": self.access_token,
                "timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                "app_key": self.app_key,
                "format": "json",
                "v": "2.0",
                "sign_method": "md5",
                "method": self.resource_method,
            }
        )
        params = dict(sorted(params.items()))
        for_sign = self.app_secret + str().join([key + value for key, value in params.items()]) + self.app_secret
        sign = md5(for_sign.encode("utf-8")).hexdigest().upper()
        params["sign"] = sign
        return params

    def get_clean_params(self, params):
        clean_params = super().get_clean_params(params=params)

        for key, value in clean_params.items():
            if (not isinstance(value, str)) and (value is not None):
                clean_params[key] = json.dumps(value)

        files = clean_params.pop("files", None)
        params_with_sign = self.add_sign(clean_params)
        params_with_sign["files"] = files  # TODO: это нужно?
        return params_with_sign

    def get_request_params(self, **kwargs) -> dict:
        clean_params = self.get_clean_params(kwargs)
        return dict(
            url=self.build_url(kwargs),
            headers=self.headers,
            data=clean_params,
        )



import re
import xml.etree.ElementTree as ET


__all__ = [
    "DictWrapper",
    "simplify_object_dict",
    "xml_to_dict",
]


class ObjectDict(dict):
    """
    Extension of dict to allow accessing keys as attributes.
    """

    def __init__(self, initd=None):
        if initd is None:
            initd = {}
        dict.__init__(self, initd)

    def __getattr__(self, item):
        node = self.__getitem__(item)

        if isinstance(node, dict) and "value" in node and len(node) == 1:
            return node["value"]
        return node

    # if value is the only key in object, you can omit it
    def __setstate__(self, item):
        return False

    def __setattr__(self, item, value):
        self.__setitem__(item, value)

    def getvalue(self, item, value=None):
        """
        Old Python 2-compatible getter method for default value.
        """
        return self.get(item, {}).get("value", value)


class XML2Dict:
    def __init__(self):
        pass

    def _parse_node(self, node):
        node_tree = ObjectDict()
        # Save attrs and text, hope there will not be a child with same name
        if node.text:
            node_tree.value = node.text
        for key, val in node.attrib.items():
            key, val = self._namespace_split(key, ObjectDict({"value": val}))
            node_tree[key] = val
        # Save childrens
        for child in list(node):
            tag, tree = self._namespace_split(child.tag, self._parse_node(child))
            if tag not in node_tree:  # the first time, so store it in dict
                node_tree[tag] = tree
                continue
            old = node_tree[tag]
            if not isinstance(old, list):
                node_tree.pop(tag)
                node_tree[tag] = [old]  # multi times, so change old dict to a list
            node_tree[tag].append(tree)  # add the new one

        return node_tree

    def _namespace_split(self, tag, value):
        """
        Split the tag '{http://cs.sfsu.edu/csc867/myscheduler}patients'
        ns = http://cs.sfsu.edu/csc867/myscheduler
        name = patients
        """
        result = re.compile(r"\{(.*)\}(.*)").search(tag)
        if result:
            value.namespace, tag = result.groups()

        return (tag, value)

    def parse(self, filename):
        """
        Parse XML file to a dict.
        """
        file_ = open(filename, "r")
        return self.fromstring(file_.read())

    def fromstring(self, str_):
        """
        Parse a string
        """
        text = ET.fromstring(str_)
        root_tag, root_tree = self._namespace_split(text.tag, self._parse_node(text))
        return ObjectDict({root_tag: root_tree})


def remove_namespace(xml):
    """
    Strips the namespace from XML document contained in a string.
    Returns the stripped string.
    """
    regex = re.compile(' xmlns(:ns2)?="[^"]+"|(ns2:)|(xml:)')
    return regex.sub("", xml)


class DictWrapper:
    def __init__(self, xml, rootkey=None):
        self.original = xml
        self.response = None
        self._rootkey = rootkey
        self._mydict = XML2Dict().fromstring(remove_namespace(xml))
        self._response_dict = self._mydict.get(list(self._mydict.keys())[0], self._mydict)

    @property
    def parsed(self):
        if self._rootkey:
            return self._response_dict.get(self._rootkey)
        return self._response_dict


def simplify_object_dict(obj):
    """
    Упрощение структуры экземляра класса ObjectDict.
    """

    if isinstance(obj, list):
        return [simplify_object_dict(obj_child) for obj_child in obj]
    elif obj is None:
        return ""
    else:
        json_obj = {}
        if (len(obj) == 1) and ("value" in obj):
            return obj.value
        for key, val in obj.items():
            if isinstance(val, str):
                if (key == "value") and (not val.strip()):
                    continue
                json_obj[key] = val
            elif isinstance(val, dict) and (len(val) == 1) and ("value" in val):
                json_obj[key] = val["value"]
            else:
                json_obj[key] = simplify_object_dict(val)
    return json_obj


def xml_to_dict(xml: str, rootkey=None):
    return simplify_object_dict(DictWrapper(xml, rootkey=rootkey).parsed)



from zs_utils.api.etsy.base_api import EtsyAPI


class GetEtsyShippingProfileList(EtsyAPI):
    """
    Docs:
    https://www.etsy.com/openapi/developers#operation/getListingShippingProfiles
    """

    http_method = "GET"
    resource_method = "shops/{shop_id}/shipping-profiles"
    required_params = ["shop_id"]


class GetEtsyShippingProfile(EtsyAPI):
    """
    Docs:
    https://www.etsy.com/openapi/developers#operation/getListingShippingProfile
    """

    http_method = "GET"
    resource_method = "shops/{shop_id}/shipping-profiles/{shipping_profile_id}"
    required_params = ["shop_id", "shipping_profile_id"]


class DeleteEtsyShippingProfile(EtsyAPI):
    """
    Docs:
    https://www.etsy.com/openapi/developers#operation/deleteListingShippingProfile
    """

    http_method = "DELETE"
    resource_method = "shops/{shop_id}/shipping-profiles/{shipping_profile_id}"
    required_params = ["shop_id", "shipping_profile_id"]


class CreateEtsyShippingProfile(EtsyAPI):
    """
    Docs:
    https://www.etsy.com/openapi/developers#operation/createListingShippingProfile
    """

    http_method = "POST"
    resource_method = "shops/{shop_id}/shipping-profiles"
    required_params = [
        "shop_id",
        "title",
        "origin_country_iso",
        "primary_cost",
        "secondary_cost",
        "min_processing_time",
        "max_processing_time",
    ]
    allowed_params = [
        "destination_country_iso",
        "destination_region",
        "origin_postal_code",
        "shipping_carrier_id",
        "mail_class",
        "min_processing_time",
        "max_processing_time",
        "min_delivery_days",
        "max_delivery_days",
    ]


class UpdateEtsyShippingProfile(EtsyAPI):
    """
    Docs:
    https://www.etsy.com/openapi/developers#operation/updateListingShippingProfile
    """

    http_method = "PUT"
    resource_method = "shops/{shop_id}/shipping-profiles/{shipping_profile_id}"
    required_params = ["shop_id", "shipping_profile_id"]
    allowed_params = [
        "title",
        "origin_country_iso",
        "origin_postal_code",
        "min_processing_time",
        "max_processing_time",
    ]



from zs_utils.api.aliexpress.base_api import AliexpressAPI


# PRODUCT
class CreateProductAPI(AliexpressAPI):
    resource_method = "aliexpress.solution.product.post"
    required_params = ["post_product_request"]


class EditProductAPI(AliexpressAPI):
    resource_method = "aliexpress.solution.product.edit"
    required_params = ["edit_product_request"]


class GetProductAPI(AliexpressAPI):
    """
    Docs: https://developers.aliexpress.com/en/doc.htm?docId=42383&docType=2
    """

    resource_method = "aliexpress.solution.product.info.get"
    required_params = ["product_id"]


class GetProductListAPI(AliexpressAPI):
    """
    Docs: https://developers.aliexpress.com/en/doc.htm?docId=42384&docType=2
    """

    resource_method = "aliexpress.solution.product.list.get"
    required_params = ["aeop_a_e_product_list_query"]


class GetSkuAttributeAPI(AliexpressAPI):
    resource_method = "aliexpress.solution.sku.attribute.query"
    allowed_params = ["query_sku_attribute_info_request"]


# В доках нет
# class EditProductSkuInventoryAPI(AliexpressAPI):
#     resource_method = "aliexpress.solution.product.sku.inventory.edit"
#     required_params = ["edit_product_sku_inventory_request"]
#
#
# class EditProductSkuPriceAPI(AliexpressAPI):
#     resource_method = "aliexpress.solution.product.sku.price.edit"
#     required_params = ["edit_product_sku_price_request"]


class DeleteProductsAPI(AliexpressAPI):
    resource_method = "aliexpress.solution.batch.product.delete"
    required_params = ["productIds"]


class UpdateBatchProductPriceAPI(AliexpressAPI):
    """
    Docs: https://developers.aliexpress.com/en/doc.htm?docId=45140&docType=2
    """

    resource_method = "aliexpress.solution.batch.product.price.update"
    required_params = ["mutiple_product_update_list"]


class UpdateBatchInventory(AliexpressAPI):
    """
    Docs: https://developers.aliexpress.com/en/doc.htm?docId=45135&docType=2
    """

    resource_method = "aliexpress.solution.batch.product.inventory.update"
    required_params = ["mutiple_product_update_list"]


class OnlineProductAPI(AliexpressAPI):
    resource_method = "aliexpress.postproduct.redefining.onlineaeproduct"
    allowed_params = ["product_ids"]


class OfflineProductAPI(AliexpressAPI):
    resource_method = "aliexpress.postproduct.redefining.offlineaeproduct"
    allowed_params = ["product_ids"]


class SubmitFeedDataAPI(AliexpressAPI):
    resource_method = "aliexpress.solution.feed.submit"
    required_params = ["operation_type", "item_list"]


class FeedQueryAPI(AliexpressAPI):
    resource_method = "aliexpress.solution.feed.query"
    allowed_params = ["job_id"]


class GetFeedListAPI(AliexpressAPI):
    resource_method = "aliexpress.solution.feed.list.get"
    allowed_params = ["current_page", "data_type", "page_size", "status"]


# PRODUCT SCHEMA
class GetProductSchemaAPI(AliexpressAPI):
    resource_method = "aliexpress.solution.product.schema.get"
    required_params = ["aliexpress_category_id"]


class PostProductBasedOnSchemaAPI(AliexpressAPI):
    resource_method = "aliexpress.solution.schema.product.instance.post"
    required_params = ["product_instance_request"]


class UpdateProductBasedOnSchemaAPI(AliexpressAPI):
    resource_method = "aliexpress.solution.schema.product.full.update"
    required_params = ["schema_full_update_request"]


# PRODUCT GROUP
class CreateProductGroupAPI(AliexpressAPI):
    resource_method = "aliexpress.postproduct.redefining.createproductgroup"
    allowed_params = ["name", "parent_id"]


class SetProductGroupAPI(AliexpressAPI):
    resource_method = "aliexpress.postproduct.redefining.setgroups"
    allowed_params = ["product_id", "group_ids"]


class GetProductGroupAPI(AliexpressAPI):
    resource_method = "aliexpress.product.productgroups.get"


# PRODUCT IMAGE
class UploadProductImageAPI(AliexpressAPI):
    resource_method = "aliexpress.photobank.redefining.uploadimageforsdk"
    allowed_params = ["group_id", "image_bytes", "file_name"]



from zs_utils.api.shipstation.base_api import ShipstationAPI


class GetShipstationRatesListAPI(ShipstationAPI):
    """
    Docs: https://www.shipstation.com/docs/api/shipments/get-rates/
    """

    http_method = "POST"
    resource_method = "shipments/getrates"
    allowed_params = [
        "carrierCode",
        "serviceCode",
        "packageCode",
        "fromPostalCode",
        "toState",
        "toCountry",
        "toPostalCode",
        "toCity",
        "weight",
        "dimensions",
        "confirmation",
        "residential",
    ]



from zs_utils.api.shopify.base_api import ShopifyAPI


class CreateShopifyImageAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/product-image#post-products-product-id-images
    """

    http_method = "POST"
    resource_method = "products/{product_id}/images.json"
    payload_key = "image"
    required_params = ["product_id", "attachment"]
    allowed_params = [
        "position",
        "variant_ids",
        "metafields",
        "alt",
    ]


class DeleteShopifyImageAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/product-image#delete-products-product-id-images-image-id
    """

    http_method = "DELETE"
    resource_method = "products/{product_id}/images/{image_id}.json"
    required_params = ["product_id", "image_id"]


class UpdateShopifyImageAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/product-image#put-products-product-id-images-image-id
    """

    http_method = "PUT"
    resource_method = "products/{product_id}/images/{image_id}.json"
    payload_key = "image"
    required_params = ["product_id", "image_id"]
    allowed_params = [
        "position",
        "alt",
        "variant_ids",
        "metafields",
    ]


class GetShopifyProductImageListAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/product-image#get-products-product-id-images
    """

    http_method = "GET"
    resource_method = "products/{product_id}/images.json"
    required_params = ["product_id"]
    allowed_params = [
        "fields",
        "since_id",
    ]


class GetShopifyProductImageAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/product-image#get-products-product-id-images-image-id
    """

    http_method = "GET"
    resource_method = "products/{product_id}/images/{image_id}.json"
    required_params = ["product_id", "image_id"]
    allowed_params = ["fields"]



from zs_utils.api.apiship.base_api import ApishipAPI


class ApishipGetOrderLabelAPI(ApishipAPI):
    resource_method = "orders/labels"
    http_method = "POST"
    required_params = ["orderIds", "format"]


class ApishipGetOrderWaybillsAPI(ApishipAPI):
    resource_method = "orders/waybills"
    http_method = "POST"
    required_params = ["orderIds"]



from zs_utils.views import CustomModelViewSet
from zs_utils.api import serializers, filters


__all__ = [
    "APIRequestLogView",
]


class APIRequestLogView(CustomModelViewSet):
    """
    View для просмотра данных модели APIRequestLog
    """

    serializer_classes = {
        "default": serializers.APIRequestLogSerializer,
        "list": serializers.APIRequestLogListSerializer,
    }
    filterset_class = filters.APIRequestLogFilter



import requests

from django.db import transaction
from django.db.models import QuerySet
from django.apps import apps
from django.conf import settings
from django.utils import timezone

from zs_utils.api.amocrm import models


__all__ = [
    "AmocrmService",
]


class AmocrmService:
    @classmethod
    def get_amocrm_app_model(cls, raise_exception: bool = True) -> type[models.AbstractAmocrmApp] | None:
        if getattr(settings, "AMOCRM_APP_MODEL", None):
            app_label, model_name = settings.AMOCRM_APP_MODEL.split(".")
            model = apps.get_model(app_label=app_label, model_name=model_name)
        else:
            model = models.AmocrmApp

        if (not model) and raise_exception:
            raise ValueError("Необходимо задать настройку 'AMOCRM_APP_MODEL'.")

        return model

    @classmethod
    def get_amocrm_app(cls, app_id: str) -> models.AbstractAmocrmApp:
        return cls.get_amocrm_app_model(raise_exception=True).objects.get(id=app_id)

    @classmethod
    def get_default_amocrm_app(cls) -> models.AbstractAmocrmApp:
        return cls.get_amocrm_app_model(raise_exception=True).objects.get(is_default=True)

    @classmethod
    def retrieve_amocrm_tokens(
        cls,
        client_id: str,
        client_secret: str,
        redirect_uri: str,
        refresh_token: str = None,
        code: str = None,
    ) -> dict:
        """
        Получение новых токенов доступа и обновления либос помощью текущего токена обновления,
        либо с помощью одноразового кода.
        Docs: https://www.amocrm.ru/developers/content/oauth/step-by-step#get_access_token
        """

        assert bool(refresh_token) != bool(code), "Необходимо задать либо 'refresh_token', либо 'code'."

        payload = {
            "client_id": client_id,
            "client_secret": client_secret,
            "redirect_uri": redirect_uri,
        }
        if refresh_token:
            payload.update({"grant_type": "refresh_token", "refresh_token": refresh_token})
        else:
            payload.update({"grant_type": "authorization_code", "code": code})

        response = requests.post(url=settings.AMOCRM_API_URL + "oauth2/access_token", json=payload)
        response.raise_for_status()

        return response.json()

    @classmethod
    @transaction.atomic()
    def refresh_amocrm_access_token(cls, app_id: str = None) -> str:
        """
        Получение токена доступа amoCRM
        """

        app_qs: QuerySet[models.AbstractAmocrmApp] = cls.get_amocrm_app_model(
            raise_exception=True
        ).objects.select_for_update()
        if app_id:
            app = app_qs.get(id=app_id)
        else:
            app = app_qs.get(is_default=True)
        if not app.refresh_token:
            raise ValueError("Отсутствует токен обновления AmoCRM.")

        # Получение новых токенов
        response_data: dict = cls.retrieve_amocrm_tokens(
            client_id=app.client_id,
            client_secret=app.client_secret,
            redirect_uri=app.redirect_uri,
            refresh_token=app.refresh_token,
        )

        # Сохранение нового refresh_token
        app.access_token = response_data["access_token"]
        app.access_token_expiry = timezone.now() + timezone.timedelta(seconds=response_data["expires_in"])
        app.refresh_token = response_data["refresh_token"]
        app.refresh_token_expiry = timezone.now() + timezone.timedelta(days=90)
        app.save()

        return app.access_token

    @classmethod
    def refresh_amocrm_apps(cls) -> None:
        if not settings.AMOCRM_APPS:
            return None

        for app_config in settings.AMOCRM_APPS:
            for key in [
                "name",
                "is_default",
                "client_id",
                "client_secret",
                "redirect_uri",
            ]:
                if app_config.get(key) is None:
                    raise ValueError(f"У конфига приложения amoCRM не задан ключ '{key}'.")

        if sum(int(app_config["is_default"]) for app_config in settings.AMOCRM_APPS) != 1:
            raise ValueError("Среди приложений amoCRM ровно одно должно быть приложением по умолчанию.")

        app_model = cls.get_amocrm_app_model(raise_exception=True)

        for app_config in settings.AMOCRM_APPS:
            app_model.objects.update_or_create(name=app_config["name"], defaults=app_config)



from zs_utils.api.ebay.base_api import EbayAPI


class CreateInventoryLocation(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/inventory/resources/location/methods/createInventoryLocation
    """

    http_method = "POST"
    resource_method = "sell/inventory/v1/location/{merchantLocationKey}"
    required_params = ["merchantLocationKey"]


class DeleteInventoryLocation(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/inventory/resources/location/methods/deleteInventoryLocation
    """

    http_method = "DELETE"
    resource_method = "sell/inventory/v1/location/{merchantLocationKey}"
    required_params = ["merchantLocationKey"]


class DisableInventoryLocation(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/inventory/resources/location/methods/disableInventoryLocation
    """

    http_method = "POST"
    resource_method = "sell/inventory/v1/location/{merchantLocationKey}/disable"
    required_params = ["merchantLocationKey"]


class EnableInventoryLocation(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/inventory/resources/location/methods/enableInventoryLocation
    """

    http_method = "POST"
    resource_method = "sell/inventory/v1/location/{merchantLocationKey}/enable"
    required_params = ["merchantLocationKey"]


class GetInventoryLocation(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/inventory/resources/location/methods/getInventoryLocation
    """

    http_method = "GET"
    resource_method = "sell/inventory/v1/location/{merchantLocationKey}"
    required_params = ["merchantLocationKey"]


class GetInventoryLocations(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/inventory/resources/location/methods/getInventoryLocations
    """

    http_method = "GET"
    resource_method = "sell/inventory/v1/location"
    allowed_params = ["offset", "limit"]


class UpdateInventoryLocation(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/inventory/resources/location/methods/updateInventoryLocation
    """

    http_method = "POST"
    resource_method = "sell/inventory/v1/location/{merchantLocationKey}/update_location_details"
    required_params = ["merchantLocationKey"]



from zs_utils.api.fedex.base_api import FedexAPI


class FedexCreatePickup(FedexAPI):
    http_method = "POST"
    resource_method = "/pickup/v1/pickups"
    required_params = [
        "associatedAccountNumber",
        "originDetail",
        "carrierCode",
    ]
    allowed_params = [
        "associatedAccountNumberType",
        "totalWeight",
        "packageCount",
        "accountAddressOfRecord",
        "remarks",
        "countryRelationships",
        "pickupType",
        "trackingNumber",
        "commodityDescription",
        "expressFreightDetail",
        "oversizePackageCount",
        "pickupNotificationDetail",
    ]


class FedexCheckPickupAvailability(FedexAPI):
    http_method = "POST"
    resource_method = "/pickup/v1/pickups/availabilities"
    required_params = [
        "pickupAddress",
        "pickupRequestType",
        "carriers",
        "countryRelationship",
    ]
    allowed_params = [
        "dispatchDate",
        "packageReadyTime",
        "customerCloseTime",
        "pickupType",
        "shipmentAttributes",
        "numberOfBusinessDays",
        "packageDetails",
        "associatedAccountNumber",
        "associatedAccountNumberType",
    ]


class FedexCancelPickup(FedexAPI):
    http_method = "PUT"
    resource_method = "/pickup/v1/pickups/cancel"
    required_params = [
        "associatedAccountNumber",
        "pickupConfirmationCode",
        "scheduledDate",
    ]
    allowed_params = [
        "remarks",
        "carrierCode",
        "accountAddressOfRecord",
        "location",
    ]



from model_utils import Choices
from zs_utils.data.enums import COUNTRIES

from django.utils.translation import gettext_lazy as _


WeightUnitOfMeasureEnum = Choices(
    ("POUND", _("Фунт")),
    ("KILOGRAM", _("Килограмм")),
    ("OUNCE", _("Унция")),
    ("GRAM", _("Грамм")),
)

LengthUnitOfMeasureEnum = Choices(
    ("INCH", _("Дюйм")),
    ("FEET", _("Фут")),
    ("CENTIMETER", _("Сантиметр")),
    ("METER", _("Метр")),
)

TimeDurationUnitEnum = Choices(
    ("YEAR", _("Год")),
    ("MONTH", _("Месяц")),
    ("DAY", _("День")),
    ("HOUR", _("Час")),
    ("CALENDAR_DAY", _("Календарный день")),
    ("BUSINESS_DAY", _("Рабочий день")),
    ("MINUTE", _("Минута")),
    ("SECOND", _("Секунда")),
    ("MILLISECOND", _("Миллисекунда")),
)

DayOfWeekEnum = Choices(
    ("MONDAY", _("Понедельник")),
    ("TUESDAY", _("Вторник")),
    ("WEDNESDAY", _("Среда")),
    ("THURSDAY", _("Четверг")),
    ("FRIDAY", _("Пятница")),
    ("SATURDAY", _("Суббота")),
    ("SUNDAY", _("Воскресенье")),
)

EbayCountryCodeEnum = COUNTRIES.subset(
    "US",
    "CA",
    "DE",
    "ES",
    "FR",
    "GB",
    "IT",
    "AU",
    "JP",
    "CN",
    "BR",
    "MX",
    "IN",
    "TR",
    "TR",
    "AE",
    "AE",
    "SG",
)



class YMLException(Exception):
    """
    Basic error for the whole module.
    """

    pass


class ValidationError(YMLException):
    """
    Data validation exception.
    """

    pass


class ParseError(YMLException):
    """
    Base parse exception.
    """

    pass



from zs_utils.api.base_api import BaseAPI


class InsalesAPI(BaseAPI):
    def __init__(self, domain: str, token: str, **kwargs):
        super().__init__(**kwargs)

        self.url = f"https://{domain}"
        self.token = token

    def validate_attrs(self) -> None:
        pass

    @property
    def production_api_url(self) -> str:
        return f"{self.url}/admin/"

    @property
    def headers(self) -> dict:
        return {
            "Authorization": f"Basic {self.token}",
            "Content-Type": "application/json",
        }



from rest_framework.request import Request
from rest_framework.permissions import BasePermission, AllowAny

from django.contrib.auth.models import AnonymousUser


__all__ = [
    "UserHasAccess", "StaffPermission"
]


class UserHasAccess(BasePermission):
    def user_has_access(self, user) -> bool:
        return True

    def has_permission(self, request: Request, view):
        if getattr(view, "ignore_has_access", False) or (AllowAny in view.permission_classes):
            return True
        return (
            bool(request.user)
            and (not isinstance(request.user, AnonymousUser))
            and self.user_has_access(user=request.user)
        )


class StaffPermission(BasePermission):
    """
    Права для сотрудников.
    """

    def has_permission(self, request: Request, view) -> bool:
        return request.user and getattr(request.user, "is_staff", False)

    def has_object_permission(self, request: Request, view, obj) -> bool:
        return self.has_permission(request=request, view=view)



from zs_utils.api.ebay.base_api import EbayAPI


class CreateFulfillmentPolicy(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/fulfillment_policy/methods/createFulfillmentPolicy
    """

    http_method = "POST"
    resource_method = "sell/account/v1/fulfillment_policy"


class DeleteFulfillmentPolicy(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/fulfillment_policy/methods/deleteFulfillmentPolicy
    """

    http_method = "DELETE"
    resource_method = "sell/account/v1/fulfillment_policy/{fulfillment_policy_id}"
    required_params = ["fulfillment_policy_id"]


class GetEbayFulfillmentPolicy(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/fulfillment_policy/methods/getFulfillmentPolicy
    """

    http_method = "GET"
    resource_method = "sell/account/v1/fulfillment_policy/{fulfillment_policy_id}"
    required_params = ["fulfillment_policy_id"]


class GetFulfillmentPolicies(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/fulfillment_policy/methods/getFulfillmentPolicies
    """

    http_method = "GET"
    resource_method = "sell/account/v1/fulfillment_policy"
    required_params = ["marketplace_id"]


class GetFulfillmentPolicyByName(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/fulfillment_policy/methods/getFulfillmentPolicyByName
    """

    http_method = "GET"
    resource_method = "sell/account/v1/fulfillment_policy/get_by_policy_name"
    required_params = ["marketplace_id", "name"]


class UpdateEbayFulfillmentPolicy(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/sell/account/resources/fulfillment_policy/methods/updateFulfillmentPolicy
    """

    http_method = "PUT"
    resource_method = "sell/account/v1/fulfillment_policy/{fulfillment_policy_id}"
    required_params = ["fulfillment_policy_id"]



import simplejson
import datetime
import decimal
import uuid
from rest_framework.renderers import JSONRenderer
from rest_framework.compat import INDENT_SEPARATORS, LONG_SEPARATORS, SHORT_SEPARATORS

from django.db.models.query import QuerySet
from django.utils import timezone
from django.utils.encoding import force_str
from django.utils.functional import Promise


class CustomJSONEncoder(simplejson.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Exception):
            return getattr(obj, "message", str(obj))

        # Код ниже взят из rest_framework.utils.encoders.JSONEncoder

        # For Date Time string spec, see ECMA 262
        # https://ecma-international.org/ecma-262/5.1/#sec-15.9.1.15
        if isinstance(obj, Promise):
            return force_str(obj)
        elif isinstance(obj, datetime.datetime):
            representation = obj.isoformat()
            # TODO: Разобраться, зачем это было нужно
            # if representation.endswith("+00:00"):
            #     representation = representation[:-6]  # + "Z"
            return representation
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, datetime.time):
            if timezone and timezone.is_aware(obj):
                raise ValueError("JSON can't represent timezone-aware times.")
            representation = obj.isoformat()
            return representation
        elif isinstance(obj, datetime.timedelta):
            return str(obj.total_seconds())
        elif isinstance(obj, decimal.Decimal):
            # Serializers will coerce decimals to strings by default.
            return float(obj)
        elif isinstance(obj, uuid.UUID):
            return str(obj)
        elif isinstance(obj, QuerySet):
            return tuple(obj)
        elif isinstance(obj, bytes):
            # Best-effort for binary blobs. See #4187.
            return obj.decode()
        elif hasattr(obj, "tolist"):
            # Numpy arrays and array scalars.
            return obj.tolist()
        elif hasattr(obj, "__getitem__"):
            cls = list if isinstance(obj, (list, tuple)) else dict
            try:
                return cls(obj)
            except Exception:
                pass
        elif hasattr(obj, "__iter__"):
            return tuple(item for item in obj)

        return super().default(obj)


def custom_json_dumps(obj, **kwargs):
    return simplejson.dumps(obj, cls=CustomJSONEncoder, **kwargs)


def custom_json_loads(obj, **kwargs):
    return simplejson.loads(obj, **kwargs)


def pretty_json(data: dict, safe: bool = True) -> str:
    try:
        return simplejson.dumps(data, indent=True, cls=CustomJSONEncoder, ensure_ascii=False)
    except TypeError as error:
        if not safe:
            raise error
    return str(data)


class CustomJSONRenderer(JSONRenderer):
    encoder_class = CustomJSONEncoder

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Render `data` into JSON, returning a bytestring.
        """
        if data is None:
            return b""

        renderer_context = renderer_context or {}
        indent = self.get_indent(accepted_media_type, renderer_context)

        if indent is None:
            separators = SHORT_SEPARATORS if self.compact else LONG_SEPARATORS
        else:
            separators = INDENT_SEPARATORS

        # Замена json на simplejson ради параметра ignore_nan
        ret = simplejson.dumps(
            data,
            cls=self.encoder_class,
            indent=indent,
            ensure_ascii=self.ensure_ascii,
            separators=separators,
            ignore_nan=True,
        )

        # We always fully escape \u2028 and \u2029 to ensure we output JSON
        # that is a strict javascript subset.
        # See: http://timelessrepo.com/json-isnt-a-javascript-subset
        ret = ret.replace("\u2028", "\\u2028").replace("\u2029", "\\u2029")
        return ret.encode()



from abc import ABC, abstractmethod
from datetime import datetime
from typing import List, Optional
import warnings

from zs_utils.yandex_market_language.exceptions import ValidationError
from .abstract import AbstractModel, XMLElement, XMLSubElement
from .price import Price, OldPrice
from .option import Option
from .outlet import Outlet
from .parameter import Parameter
from .condition import Condition
from .dimensions import Dimensions
from .age import Age
from . import fields


EXPIRY_FORMAT = "YYYY-MM-DDThh:mm"
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class AbstractOffer(
    fields.OutletsField,
    fields.EnableAutoDiscountField,
    fields.DeliveryOptionsField,
    fields.PickupOptionsField,
    AbstractModel,
    ABC,
):
    """
    Abstract offer model for all other offer models.
    """

    __TYPE__ = None

    __slots__ = [
        "_cbid",
        "_delivery",
        "_pickup",
        "_store",
        "_min_quantity",
        "_manufacturer_warranty",
        "_adult",
        "_parameters",
        "_expiry",
        "_weight",
        "_downloadable",
        "_available",
        "_group_id",
        "offer_id",
        "url",
        "price",
        "currency",
        "category_id",
        "vendor",
        "vendor_code",
        "bid",
        "old_price",
        "_enable_auto_discounts",
        "pictures",
        "supplier",
        "_delivery_options",
        "_outlets",
        "_pickup_options",
        "description",
        "sales_notes",
        "barcodes",
        "condition",
        "credit_template_id",
        "dimensions",
        "age",
    ]

    def __init__(
        self,
        offer_id=None,
        price: Price = None,
        old_price: OldPrice = None,
        currency: str = "RUR",
        category_id=None,
        url=None,
        vendor=None,
        vendor_code=None,
        bid=None,
        cbid=None,
        enable_auto_discounts=None,
        pictures: List[str] = None,
        supplier=None,
        delivery=True,
        pickup=True,
        delivery_options: List[Option] = None,
        outlets: List[Option] = None,
        pickup_options: List[Option] = None,
        store=None,
        description: str = None,
        sales_notes: str = None,
        min_quantity: int = None,
        manufacturer_warranty=None,
        country_of_origin=None,
        adult=None,
        barcodes: List[dict] = None,
        parameters: List[Parameter] = None,
        condition: Condition = None,
        credit_template_id: str = None,
        expiry=None,
        weight=None,
        dimensions: Dimensions = None,
        downloadable=None,
        available=None,
        age: Age = None,
        group_id=None,
        **kwargs,
    ):
        self.vendor = vendor
        self.vendor_code = vendor_code
        self.offer_id = offer_id
        self.bid = bid
        self.cbid = cbid
        self.url = url
        self.price = price
        self.old_price = old_price
        self.enable_auto_discounts = enable_auto_discounts
        self.currency = currency
        self.category_id = category_id
        self.pictures = pictures
        self.supplier = supplier
        self.delivery = delivery
        self.pickup = pickup
        self.delivery_options = delivery_options
        self.outlets = outlets
        self.pickup_options = pickup_options
        self.store = store
        self.description = description
        self.sales_notes = sales_notes
        self.min_quantity = min_quantity
        self.manufacturer_warranty = manufacturer_warranty
        self.country_of_origin = country_of_origin
        self.adult = adult
        self.barcodes = barcodes
        self.parameters = parameters
        self.condition = condition
        self.credit_template_id = credit_template_id
        self.expiry = expiry
        self.weight = weight
        self.dimensions = dimensions
        self.downloadable = downloadable
        self.available = available
        self.age = age
        self.group_id = group_id

    @property
    def cbid(self):
        return self._cbid

    @cbid.setter
    def cbid(self, value):
        if value:
            warnings.warn("The attribute cbid is deprecated. " "Use the attribute bid instead.", DeprecationWarning)
        self._cbid = value

    @property
    def delivery(self) -> bool:
        return True if self._delivery == "true" else False

    @delivery.setter
    def delivery(self, value):
        if value is None:
            value = True
        self._delivery = self._is_valid_bool(value, "delivery")

    @property
    def pickup(self) -> bool:
        return True if self._delivery == "true" else False

    @pickup.setter
    def pickup(self, value):
        if value is None:  # sets default True value for pickup
            value = True
        self._pickup = self._is_valid_bool(value, "pickup")

    @property
    def store(self) -> Optional[bool]:
        return self._str_to_bool(self._store)

    @store.setter
    def store(self, value):
        self._store = self._is_valid_bool(value, "store", True)

    @property
    def min_quantity(self) -> int:
        return int(self._min_quantity) if self._min_quantity else None

    @min_quantity.setter
    def min_quantity(self, value):
        if value is None:  # sets default value for min_quantity
            self._min_quantity = None
        else:
            self._min_quantity = self._is_valid_int(value, "min_quantity")

    @property
    def manufacturer_warranty(self) -> Optional[bool]:
        return self._str_to_bool(self._manufacturer_warranty)

    @manufacturer_warranty.setter
    def manufacturer_warranty(self, value):
        self._manufacturer_warranty = self._is_valid_bool(value, "manufacturer_warranty", True)

    @property
    def adult(self) -> Optional[bool]:
        return self._str_to_bool(self._adult)

    @adult.setter
    def adult(self, value):
        self._adult = self._is_valid_bool(value, "adult", True)

    @property
    def parameters(self) -> List[Parameter]:
        return self._parameters or []

    @parameters.setter
    def parameters(self, params):
        self._parameters = params or []

    @property
    def expiry(self) -> Optional[datetime]:
        if self._expiry:
            return datetime.strptime(self._expiry, EXPIRY_FORMAT)
        else:
            return None

    @expiry.setter
    def expiry(self, dt):
        self._expiry = self._is_valid_datetime(dt, EXPIRY_FORMAT, "expiry", True)

    @property
    def weight(self) -> Optional[float]:
        return float(self._weight) if self._weight else None

    @weight.setter
    def weight(self, value):
        if isinstance(value, str):
            value = value.replace(",", ".")
        elif not value:
            value = None
        self._weight = self._is_valid_float(value, "weight", True)

    @property
    def downloadable(self) -> Optional[bool]:
        return self._str_to_bool(self._downloadable)

    @downloadable.setter
    def downloadable(self, value):
        self._downloadable = self._is_valid_bool(value, "downloadable", True)

    @property
    def available(self) -> Optional[bool]:
        return self._str_to_bool(self._available)

    @available.setter
    def available(self, value):
        self._available = self._is_valid_bool(value, "available", True)

    @property
    def group_id(self) -> Optional[str]:
        return str(self._group_id) if self._group_id else None

    @group_id.setter
    def group_id(self, value):
        # value = self._is_valid_int(value, "group_id", True)

        # Validate group id and raise an error if it's not valid
        # if len(str(value)) > 9:
        #     raise ValidationError("group_id must be an integer, maximum 9 characters.")

        self._group_id = str(value) if value else None

    @abstractmethod
    def create_dict(self, **kwargs) -> dict:
        return dict(
            type=self.__TYPE__,
            vendor=self.vendor,
            vendor_code=self.vendor_code,
            offer_id=self.offer_id,
            bid=self.bid,
            url=self.url,
            price=self.price.to_dict() if self.price else None,
            old_price=self.old_price,
            enable_auto_discounts=self.enable_auto_discounts,
            currency=self.currency,
            category_id=self.category_id,
            pictures=self.pictures,
            supplier=self.supplier,
            delivery=self.delivery,
            pickup=self.pickup,
            delivery_options=[o.to_dict() for o in self.delivery_options],
            outlets=[o.to_dict() for o in self.outlets],
            pickup_options=[o.to_dict() for o in self.pickup_options],
            store=self.store,
            description=self.description,
            sales_notes=self.sales_notes,
            min_quantity=self.min_quantity,
            manufacturer_warranty=self.manufacturer_warranty,
            country_of_origin=self.country_of_origin,
            adult=self.adult,
            barcodes=self.barcodes,
            parameters=[p.to_dict() for p in self.parameters],
            condition=self.condition.to_dict() if self.condition else None,
            credit_template_id=self.credit_template_id,
            expiry=self.expiry,
            weight=self.weight,
            dimensions=self.dimensions.to_dict() if self.dimensions else None,
            downloadable=self.downloadable,
            available=self.available,
            age=self.age.to_dict() if self.age else None,
            group_id=self.group_id,
            **kwargs,
        )

    @abstractmethod
    def create_xml(self, **kwargs) -> XMLElement:
        offer_el = XMLElement("offer", {"id": self.offer_id})

        # Add offer type
        if self.__TYPE__:
            offer_el.attrib["type"] = self.__TYPE__

        # Add offer bid attribute
        if self.bid:
            offer_el.attrib["bid"] = str(self.bid)

        # Add offer cbid attribute (deprecated)
        if self.cbid:
            offer_el.attrib["cbid"] = self.cbid

        # Add offer available attribute
        if self.available is not None:
            offer_el.attrib["available"] = self._available

        if self.group_id:
            offer_el.attrib["group_id"] = self._group_id

        # Add simple values
        for tag, attr in {
            "vendor": "vendor",
            "vendorCode": "vendor_code",
            "url": "url",
            "enable_auto_discounts": "_enable_auto_discounts",
            "currencyId": "currency",
            "categoryId": "category_id",
            "delivery": "_delivery",
            "pickup": "_pickup",
            "store": "_store",
            "description": "description",
            "sales_notes": "sales_notes",
            "min-quantity": "_min_quantity",
            "manufacturer_warranty": "_manufacturer_warranty",
            "country_of_origin": "country_of_origin",
            "adult": "_adult",
            "expiry": "_expiry",
            "weight": "_weight",
            "downloadable": "_downloadable",
            **kwargs,
        }.items():
            value = getattr(self, attr)
            if value:
                el = XMLSubElement(offer_el, tag)
                el.text = value

        # Add price
        self.price.to_xml(offer_el)
        if self.old_price:
            self.old_price.to_xml(offer_el)

        # Add pictures
        if self.pictures:
            for url in self.pictures:
                picture_el = XMLSubElement(offer_el, "picture")
                picture_el.text = url

        # Add supplier
        if self.supplier:
            XMLSubElement(offer_el, "supplier", {"ogrn": self.supplier})

        # Add delivery options
        if self.delivery_options:
            delivery_options_el = XMLSubElement(offer_el, "delivery-options")
            for o in self.delivery_options:
                o.to_xml(delivery_options_el)

        # Add pickup options
        if self.pickup_options:
            pickup_options_el = XMLSubElement(offer_el, "pickup-options")
            for o in self.pickup_options:
                o.to_xml(pickup_options_el)

        # Add outlets
        if self.outlets:
            outlets_el = XMLSubElement(offer_el, "outlets")
            for o in self.outlets:
                o.to_xml(outlets_el)

        # Add barcodes
        if self.barcodes:
            for b in self.barcodes:
                b_el = XMLSubElement(offer_el, "barcode")
                b_el.text = b["value"]

        # Add parameters
        if self.parameters:
            for p in self.parameters:
                p.to_xml(offer_el)

        # Add condition
        if self.condition:
            self.condition.to_xml(offer_el)

        # Add credit template
        if self.credit_template_id:
            XMLSubElement(offer_el, "credit-template", {"id": self.credit_template_id})

        # Add dimensions
        if self.dimensions:
            self.dimensions.to_xml(offer_el)

        # Add age
        if self.age:
            self.age.to_xml(offer_el)

        return offer_el

    @staticmethod
    @abstractmethod
    def from_xml(offer_el: XMLElement, **mapping) -> dict:
        """
        Abstract method for parsing the xml element into a dictionary.
        """
        kwargs = {}
        mapping = {
            "vendorCode": "vendor_code",
            "oldprice": "old_price",
            "currencyId": "currency",
            "categoryId": "category_id",
            "min-quantity": "min_quantity",
            **mapping,
        }

        pictures = []
        barcodes = []
        parameters = []

        for el in offer_el:
            if el.tag == "picture":
                pictures.append(el.text)
            elif el.tag == "delivery-options":
                delivery_options = []
                for option_el in el:
                    delivery_options.append(Option.from_xml(option_el))
                kwargs["delivery_options"] = delivery_options
            elif el.tag == "outlets":
                outlets = []
                for outlet_el in el:
                    outlets.append(Outlet.from_xml(outlet_el))
                kwargs["outlets"] = outlets
            elif el.tag == "pickup-options":
                pickup_options = []
                for option_el in el:
                    pickup_options.append(Option.from_xml(option_el))
                kwargs["pickup_options"] = pickup_options
            elif el.tag == "barcode":
                barcodes.append({"value": el.text, "type": el.attrib.get("type")})
            elif el.tag == "param":
                parameters.append(Parameter.from_xml(el))
            elif el.tag == "credit-template":
                kwargs["credit_template_id"] = el.attrib["id"]
            elif el.tag == "dimensions":
                if el.text:
                    kwargs["dimensions"] = Dimensions.from_xml(el)
                else:
                    kwargs["dimensions"] = None
            elif el.tag == "price":
                kwargs["price"] = Price.from_xml(el)
            elif el.tag == "condition":
                kwargs["condition"] = Condition.from_xml(el)
            elif el.tag == "age":
                kwargs["age"] = Age.from_xml(el)
            elif el.tag == "supplier":
                kwargs["supplier"] = el.attrib["ogrn"]
            else:
                if (el.tag == "oldprice") and isinstance(el.text, str):
                    el.text = el.text.replace(",", ".")

                k = mapping.get(el.tag, el.tag)
                kwargs[k] = el.text

        if pictures:
            kwargs["pictures"] = pictures
        if barcodes:
            kwargs["barcodes"] = barcodes
        if parameters:
            kwargs["parameters"] = parameters

        kwargs["offer_id"] = offer_el.attrib["id"]
        kwargs["group_id"] = offer_el.attrib.get("group_id")
        kwargs["bid"] = offer_el.attrib.get("bid")
        kwargs["cbid"] = offer_el.attrib.get("cbid")
        kwargs["available"] = offer_el.attrib.get("available")

        return kwargs


class SimplifiedOffer(AbstractOffer):
    """
    Simplified offer.
    In a simplified type, the manufacturer, type and name of the goods
    are indicated in one element - name.

    Yandex.Market docs:
    https://yandex.ru/support/partnermarket/offers.html
    """

    __TYPE__ = None

    __slots__ = [*AbstractOffer.__slots__, "name"]

    def __init__(self, name: str = "", **kwargs):
        super().__init__(**kwargs)
        self.name = name

    def create_dict(self, **kwargs) -> dict:
        return super().create_dict(name=self.name)

    def create_xml(self, **kwargs) -> XMLElement:
        offer_el = super().create_xml()
        name_el = XMLElement("name")
        name_el.text = self.name
        offer_el.insert(0, name_el)
        return offer_el

    @staticmethod
    def from_xml(offer_el: XMLElement, **mapping) -> "SimplifiedOffer":
        kwargs = AbstractOffer.from_xml(offer_el, **mapping)
        return SimplifiedOffer(**kwargs)


class ArbitraryOffer(AbstractOffer):
    """
    Arbitrary offer.
    In an arbitrary type, the manufacturer, type and name of the product
    must be indicated in separate elements - model, vendor & typePrefix.

    Yandex.Market docs:
    https://yandex.ru/support/partnermarket/export/vendor-model.html
    """

    __TYPE__ = "vendor.model"

    __slots__ = [*AbstractOffer.__slots__, "model", "vendor", "type_prefix"]

    def __init__(self, vendor: str = None, model: str = None, type_prefix: str = None, **kwargs):
        super().__init__(vendor=vendor, **kwargs)
        self.model = model
        self.type_prefix = type_prefix

    def create_dict(self, **kwargs) -> dict:
        return super().create_dict(model=self.model, type_prefix=self.type_prefix)

    def create_xml(self, **kwargs) -> XMLElement:
        offer_el = super().create_xml()

        # Add model element
        model_el = XMLSubElement(offer_el, "model")
        model_el.text = self.model

        # Add typePrefix element
        if self.type_prefix:
            type_prefix_el = XMLSubElement(offer_el, "typePrefix")
            type_prefix_el.text = self.type_prefix

        return offer_el

    @staticmethod
    def from_xml(offer_el: XMLElement, **mapping) -> "ArbitraryOffer":
        mapping.update(
            {
                "typePrefix": "type_prefix",
            }
        )
        kwargs = AbstractOffer.from_xml(offer_el, **mapping)
        return ArbitraryOffer(**kwargs)


class AbstractBookOffer(fields.YearField, AbstractOffer, ABC):
    """
    Abstract book offer for book & audio book offer types.
    """

    __slots__ = [
        *AbstractOffer.__slots__,
        "_volume",
        "_part",
        "name",
        "publisher",
        "age",
        "isbn",
        "author",
        "series",
        "_year",
        "language",
        "table_of_contents",
    ]

    def __init__(
        self,
        name: str,
        publisher: str,
        age: Age,
        isbn: str = None,
        author: str = None,
        series: str = None,
        year=None,
        volume=None,
        part=None,
        language: str = None,
        table_of_contents=None,
        **kwargs,
    ):
        super().__init__(age=age, **kwargs)
        self.name = name
        self.publisher = publisher
        self.isbn = isbn
        self.author = author
        self.series = series
        self.year = year
        self.volume = volume
        self.part = part
        self.language = language
        self.table_of_contents = table_of_contents

    @property
    def volume(self) -> Optional[int]:
        return int(self._volume) if self._volume else None

    @volume.setter
    def volume(self, value):
        self._volume = self._is_valid_int(value, "volume", True)

    @property
    def part(self) -> Optional[int]:
        return int(self._part) if self._part else None

    @part.setter
    def part(self, value):
        self._part = self._is_valid_int(value, "part", True)

    @abstractmethod
    def create_dict(self, **kwargs) -> dict:
        return super().create_dict(
            name=self.name,
            publisher=self.publisher,
            isbn=self.isbn,
            author=self.author,
            series=self.series,
            year=self.year,
            volume=self.volume,
            part=self.part,
            language=self.language,
            table_of_contents=self.table_of_contents,
            **kwargs,
        )

    @abstractmethod
    def create_xml(self, **kwargs) -> XMLElement:
        offer_el = super().create_xml(
            name="name",
            publisher="publisher",
            ISBN="isbn",
            author="author",
            series="series",
            year="_year",
            volume="_volume",
            part="_part",
            language="language",
            table_of_contents="table_of_contents",
            **kwargs,
        )

        return offer_el

    @staticmethod
    @abstractmethod
    def from_xml(offer_el: XMLElement, **mapping) -> dict:
        mapping.update(
            {
                "publisher": "publisher",
                "ISBN": "isbn",
            }
        )
        return AbstractOffer.from_xml(offer_el, **mapping)


class BookOffer(AbstractBookOffer):
    """
    Special offer type for books.

    Yandex.Market docs:
    https://yandex.ru/support/partnermarket/export/books.html
    """

    __TYPE__ = "book"

    __slots__ = [*AbstractBookOffer.__slots__, "binding", "_page_extent"]

    def __init__(self, binding=None, page_extent=None, **kwargs):
        super().__init__(**kwargs)
        self.binding = binding
        self.page_extent = page_extent

    @property
    def page_extent(self) -> Optional[int]:
        return int(self._page_extent) if self._page_extent else None

    @page_extent.setter
    def page_extent(self, value):
        value = self._is_valid_int(value, "page_extent", True, False)
        if value <= 0:
            raise ValidationError("page_extent must be positive int")
        self._page_extent = str(value)

    def create_dict(self, **kwargs) -> dict:
        return super().create_dict(binding=self.binding, page_extent=self.page_extent)

    def create_xml(self, **kwargs) -> XMLElement:
        offer_el = super().create_xml(binding="binding", page_extent="_page_extent")
        return offer_el

    @staticmethod
    def from_xml(offer_el: XMLElement, **mapping) -> "BookOffer":
        kwargs = AbstractBookOffer.from_xml(offer_el, **mapping)
        return BookOffer(**kwargs)


class AudioBookOffer(AbstractBookOffer):
    """
    Audio book offer.

    Docs:
    https://yandex.ru/support/partnermarket/export/audiobooks.html
    """

    __TYPE__ = "audiobook"

    __slots__ = [
        *AbstractBookOffer.__slots__,
        "performed_by",
        "performance_type",
        "storage",
        "audio_format",
        "recording_length",
    ]

    def __init__(
        self,
        performed_by: str = None,
        performance_type: str = None,
        storage: str = None,
        audio_format: str = None,
        recording_length: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.performed_by = performed_by
        self.performance_type = performance_type
        self.storage = storage
        self.audio_format = audio_format
        self.recording_length = recording_length

    def create_dict(self, **kwargs) -> dict:
        return super().create_dict(
            performed_by=self.performed_by,
            performance_type=self.performance_type,
            storage=self.storage,
            audio_format=self.audio_format,
            recording_length=self.recording_length,
        )

    def create_xml(self, **kwargs) -> XMLElement:
        return super().create_xml(
            performed_by="performed_by",
            performance_type="performance_type",
            storage="storage",
            format="audio_format",
            recording_length="recording_length",
        )

    @staticmethod
    def from_xml(offer_el: XMLElement, **mapping) -> "AudioBookOffer":
        mapping.update({"format": "audio_format"})
        kwargs = AbstractBookOffer.from_xml(offer_el, **mapping)
        return AudioBookOffer(**kwargs)


class MusicVideoOffer(fields.YearField, AbstractOffer):
    """
    Music or video offer.

    Docs:
    https://yandex.ru/support/partnermarket/export/music-video.html
    """

    __TYPE__ = "artist.title"

    __slots__ = [
        *AbstractOffer.__slots__,
        "title",
        "artist",
        "_year",
        "media",
        "starring",
        "director",
        "original_name",
        "country",
    ]

    def __init__(
        self,
        title: str,
        artist: str = None,
        year=None,
        media: str = None,
        starring: str = None,
        director: str = None,
        original_name: str = None,
        country: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.artist = artist
        self.title = title
        self.year = year
        self.media = media
        self.starring = starring
        self.director = director
        self.original_name = original_name
        self.country = country

    def create_dict(self, **kwargs) -> dict:
        return super().create_dict(
            artist=self.artist,
            title=self.title,
            year=self.year,
            media=self.media,
            starring=self.starring,
            director=self.director,
            original_name=self.original_name,
            country=self.country,
        )

    def create_xml(self, **kwargs) -> XMLElement:
        return super().create_xml(
            artist="artist",
            title="title",
            year="_year",
            media="media",
            starring="starring",
            director="director",
            originalName="original_name",
            country="country",
        )

    @staticmethod
    def from_xml(offer_el: XMLElement, **mapping) -> "MusicVideoOffer":
        mapping.update({"originalName": "original_name"})
        kwargs = AbstractOffer.from_xml(offer_el, **mapping)
        return MusicVideoOffer(**kwargs)


class MedicineOffer(AbstractOffer):
    """
    Medicine offer.

    Docs:
    https://yandex.ru/support/partnermarket/export/medicine.html
    """

    __TYPE__ = "medicine"

    __slots__ = [*AbstractOffer.__slots__, "name", "_delivery", "_pickup"]

    def __init__(self, name, delivery, pickup, **kwargs):
        super().__init__(delivery=delivery, pickup=pickup, **kwargs)
        self.name = name

    def create_dict(self, **kwargs) -> dict:
        return super().create_dict(name=self.name)

    def create_xml(self, **kwargs) -> XMLElement:
        return super().create_xml(name="name")

    @staticmethod
    def from_xml(offer_el: XMLElement, **mapping) -> "MedicineOffer":
        kwargs = AbstractOffer.from_xml(offer_el)
        return MedicineOffer(**kwargs)


class EventTicketOffer(AbstractOffer):
    """
    EventTicket offer.

    Docs:
    https://yandex.ru/support/partnermarket/export/event-tickets.html
    """

    __TYPE__ = "event-ticket"

    __slots__ = [*AbstractOffer.__slots__, "_date", "_is_premiere", "_is_kids", "name", "place", "hall", "hall_part"]

    def __init__(
        self,
        name: str,
        place: str,
        date,
        hall: str = None,
        hall_part: str = None,
        is_premiere=None,
        is_kids=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.name = name
        self.place = place
        self.hall = hall
        self.hall_part = hall_part
        self.date = date
        self.is_premiere = is_premiere
        self.is_kids = is_kids

    @property
    def date(self) -> datetime:
        return datetime.strptime(self._date, DATE_FORMAT)

    @date.setter
    def date(self, dt):
        self._date = self._is_valid_datetime(dt, DATE_FORMAT, "date")

    @property
    def is_premiere(self) -> Optional[bool]:
        return self._str_to_bool(self._is_premiere)

    @is_premiere.setter
    def is_premiere(self, value):
        self._is_premiere = self._is_valid_bool(value, "is_premiere", True)

    @property
    def is_kids(self) -> Optional[bool]:
        return self._str_to_bool(self._is_kids)

    @is_kids.setter
    def is_kids(self, value):
        self._is_kids = self._is_valid_bool(value, "is_kids", True)

    def create_dict(self, **kwargs) -> dict:
        return super().create_dict(
            name=self.name,
            place=self.place,
            hall=self.hall,
            hall_part=self.hall_part,
            date=self.date,
            is_premiere=self.is_premiere,
            is_kids=self.is_kids,
        )

    def create_xml(self, **kwargs) -> XMLElement:
        return super().create_xml(
            name="name",
            place="place",
            hall="hall",
            hall_part="hall_part",
            date="_date",
            is_premiere="_is_premiere",
            is_kids="_is_kids",
        )

    @staticmethod
    def from_xml(offer_el: XMLElement, **mapping) -> "EventTicketOffer":
        kwargs = AbstractOffer.from_xml(offer_el)
        return EventTicketOffer(**kwargs)


class AlcoholOffer(AbstractOffer):
    """
    Alcohol offer.

    Docs:
    https://yandex.ru/support/partnermarket/export/alcohol.html
    """

    __TYPE__ = "alco"

    __slots__ = [*AbstractOffer.__slots__, "name", "vendor", "barcodes", "parameters"]

    def __init__(self, name: str, vendor: str, barcodes: List[str], parameters: List[Parameter], **kwargs):
        super().__init__(vendor=vendor, barcodes=barcodes, parameters=parameters, **kwargs)
        self.name = name

    def create_dict(self, **kwargs) -> dict:
        return super().create_dict(name=self.name)

    def create_xml(self, **kwargs) -> XMLElement:
        return super().create_xml(name="name")

    @staticmethod
    def from_xml(offer_el: XMLElement, **mapping) -> "AlcoholOffer":
        kwargs = AbstractOffer.from_xml(offer_el)
        return AlcoholOffer(**kwargs)



from typing import Optional


__all__ = [
    "YearField",
]


class YearField:
    _year = None

    @property
    def year(self) -> Optional[int]:
        return int(self._year) if self._year else None

    @year.setter
    def year(self, value):
        self._year = self._is_valid_int(value, "year", True)



from zs_utils.api.apiship.base_api import ApishipAPI


class ApishipCreateOrderAPI(ApishipAPI):
    """
    Docs: https://docs.apiship.ru/docs/api/order-service/create-order-async/
    """

    resource_method = "orders"
    http_method = "POST"
    required_params = [
        "order",
        "cost",
        "sender",
        "recipient",
        "places",
    ]
    allowed_params = [
        "returnAddress",
        "extraParams",
    ]


class ApishipValidateOrderAPI(ApishipAPI):
    resource_method = "orders/validate"
    http_method = "POST"
    required_params = [
        "order",
        "cost",
        "sender",
        "recipient",
        "places",
    ]
    allowed_params = [
        "returnAddress",
        "extraParams",
    ]


class ApishipGetOrderAPI(ApishipAPI):
    resource_method = "orders/{order_id}"
    http_method = "GET"
    required_params = ["order_id"]


class ApishipUpdateOrderAPI(ApishipAPI):
    resource_method = "orders/{order_id}"
    http_method = "PUT"
    required_params = [
        "order_id",
        "order",
        "cost",
        "sender",
        "recipient",
        "returnAddress",
        "places",
        "extraParams",
    ]


class ApishipDeleteOrderAPI(ApishipAPI):
    resource_method = "orders/{order_id}"
    http_method = "DELETE"
    required_params = ["order_id"]


class ApishipResendOrderAPI(ApishipAPI):
    resource_method = "orders/{order_id}/resend"
    http_method = "POST"
    required_params = ["order_id"]


class ApishipCancelOrderAPI(ApishipAPI):
    resource_method = "orders/{order_id}/cancel"
    http_method = "GET"
    required_params = ["order_id"]


class ApishipUpdateOrderItemsAPI(ApishipAPI):
    resource_method = "orders/{order_id}/items"
    http_method = "POST"
    required_params = ["order_id", "items"]



__all__ = [
    "DeliveryOptionsField",
    "PickupOptionsField",
]


class DeliveryOptionsField:
    _delivery_options = []

    @property
    def delivery_options(self):
        return self._delivery_options

    @delivery_options.setter
    def delivery_options(self, options):
        self._delivery_options = options if options else []


class PickupOptionsField:
    _pickup_options = []

    @property
    def pickup_options(self):
        return self._pickup_options

    @pickup_options.setter
    def pickup_options(self, options):
        self._pickup_options = options if options else []



from model_utils.models import UUIDModel

from django.db import models
from django.utils import timezone


__all__ = [
    "AbstractAmocrmApp",
    "AmocrmApp",
]


class AbstractAmocrmApp(UUIDModel):
    client_id = models.TextField(verbose_name="ID приложения")
    client_secret = models.TextField(verbose_name="Секретный ключ приложения")
    redirect_uri = models.TextField(verbose_name="Redirect URI приложения")

    name = models.CharField(max_length=64, verbose_name="Локальный идентификатор")
    access_token = models.TextField(null=True, verbose_name="Токен доступа")
    access_token_expiry = models.DateTimeField(null=True, verbose_name="Дата истечения токена доступа")
    refresh_token = models.TextField(null=True, verbose_name="Токен для обновления токена доступа")
    refresh_token_expiry = models.DateTimeField(null=True, verbose_name="Дата истечения токена обновления")
    is_default = models.BooleanField(default=False, verbose_name="Приложение по умолчанию")

    class Meta:
        abstract = True

    @property
    def access_token_expired(self) -> bool:
        if self.access_token_expiry:
            return self.access_token_expiry < timezone.now() + timezone.timedelta(minutes=10)
        return False

    @property
    def refresh_token_expired(self) -> bool:
        if self.refresh_token_expiry:
            return self.refresh_token_expiry < timezone.now() + timezone.timedelta(minutes=10)
        return False


class AmocrmApp(AbstractAmocrmApp):
    pass



from zs_utils.api.shipstation.base_api import ShipstationAPI


class GetShipstationCarrierListAPI(ShipstationAPI):
    http_method = "GET"
    resource_method = "carriers"


class GetShipstationCarrierAPI(ShipstationAPI):
    http_method = "GET"
    resource_method = "carriers/getcarrier"
    required_params = ["carrierCode"]



from django.utils.translation import gettext_lazy as _

from model_utils import Choices


__all__ = [
    "WEIGHT_UNITS",
    "ENGLISH_WEIGHT_UNITS",
    "METRIC_WEIGHT_UNITS",
    "LENGTH_UNITS",
    "ENGLISH_LENGTH_UNITS",
    "METRIC_LENGTH_UNITS",
]


WEIGHT_UNITS = Choices(
    ("G", _("Грамм")),
    ("KG", _("Килограмм")),
    ("LB", _("Фунт")),
    ("OZ", _("Унция")),
)

ENGLISH_WEIGHT_UNITS = WEIGHT_UNITS.subset("LB", "OZ")

METRIC_WEIGHT_UNITS = WEIGHT_UNITS.subset("G", "KG")

LENGTH_UNITS = Choices(
    ("IN", _("Дюйм")),
    ("FT", _("Фут")),
    ("YD", _("Ярд")),
    ("MM", _("Миллиметр")),
    ("CM", _("Сантиметр")),
    ("M", _("Метр")),
)

ENGLISH_LENGTH_UNITS = LENGTH_UNITS.subset("IN", "FT", "YD")

METRIC_LENGTH_UNITS = LENGTH_UNITS.subset("MM", "CM", "M")



import logging
from rest_framework.views import APIView
from rest_framework.viewsets import ModelViewSet
from rest_framework.decorators import action
from rest_framework.exceptions import MethodNotAllowed, ValidationError
from rest_framework.permissions import AllowAny
from django.conf import settings
from django.http import QueryDict
from django.utils.translation import gettext as _

from zs_utils.permissions import UserHasAccess
from zs_utils.views import mixins
from zs_utils.captcha import validate_captcha


__all__ = [
    "CustomAPIView",
    "CustomModelViewSet",
    "get_client_ip",
]


logger = logging.getLogger()


def get_client_ip(request):
    """
    Получение IP адреса клиента
    """
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[0]
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip


class CustomAPIView(mixins.DataValidationViewMixin, mixins.ResponseShortcutsViewMixin, APIView):
    permission_classes = [AllowAny]


class CustomModelViewSet(
    mixins.AdminModeViewMixin,
    mixins.DataValidationViewMixin,
    mixins.ResponseShortcutsViewMixin,
    ModelViewSet,
):
    lookup_field = "id"
    default_permission = UserHasAccess
    ignore_has_access = False
    allow_partial_update = True

    not_allowed_actions = []
    extra_allowed_actions = []

    required_filters = []
    actions_for_filterset_class = [
        "retrieve",
    ]

    serializer_classes = {}
    serializer_class = None

    def initialize_request(self, request, *args, **kwargs):
        request = super().initialize_request(request, *args, **kwargs)

        if (self.action in self.not_allowed_actions) and (self.action not in self.extra_allowed_actions):
            return self.http_method_not_allowed(request, *args, **kwargs)

        return request

    def get_permissions(self):
        """
        Проверка прав пользователя
        """
        return super().get_permissions() + [self.default_permission()]

    def get_exception_handler_context(self) -> dict:
        """
        Формирование данных для контекста ошибки.
        """

        context = super().get_exception_handler_context()

        try:
            context["user"] = self.get_user()
        except Exception as error:
            logger.error(msg=str(error))

        return context

    def get_serializer_class(self):
        """
        Получение класса сериализатора
        """

        # Маппинг между экшенами и сериализаторами
        return self.serializer_classes.get(self.action, self.serializer_classes.get("default", self.serializer_class))

    def get_serializer(self, *args, **kwargs):
        """
        Получение сериализатора
        """
        # Если данные передаются на прямую из request.data, то это объект QueryDict, он не изменяем, нужно копировать
        if ("data" in kwargs) and isinstance(kwargs["data"], QueryDict):
            kwargs["data"] = kwargs["data"].copy()

        serializer_class = self.get_serializer_class()
        kwargs.setdefault("context", self.get_serializer_context())

        return serializer_class(*args, **kwargs)

    def validate_filters(self) -> None:
        """
        Валидация фильтров запросов
        """
        for key in self.required_filters:
            if not self.request.GET.get(key):
                raise ValidationError({key: _("Обязательный фильтр")})

        if self.request.GET.get("limit") and (int(self.request.GET["limit"]) > settings.DRF_LIMIT_FILTER_MAX_VALUE):
            raise ValidationError(
                {"limit": _("Максимальное значение: {max_value}").format(max_value=settings.DRF_LIMIT_FILTER_MAX_VALUE)}
            )

    def get_queryset_filter_kwargs(self) -> dict:
        return {}

    def limit_queryset(self, queryset):
        """
        Валидация фильтров при получении списка объектов
        """
        if self.action == "list":
            self.validate_filters()

        return queryset.filter(**self.get_queryset_filter_kwargs())

    def get_queryset(self, manager: str = "objects"):
        """
        Получение Queryset
        """
        if getattr(self, "filterset_class", None) and hasattr(self.filterset_class, "Meta"):
            model = self.filterset_class.Meta.model
        else:
            model = self.get_serializer_class().Meta.model

        queryset = getattr(model, manager).all()

        if (self.actions_for_filterset_class == "__all__") or (self.action in self.actions_for_filterset_class):
            queryset = self.filter_queryset(queryset=queryset)

        if not self.no_limit:
            # Обязательная фильтрация результатов для рядовых пользователей
            queryset = self.limit_queryset(queryset=queryset)

        return queryset

    @action(detail=False, methods=["GET"])
    def count(self, request, *args, **kwargs):
        """
        Получение кол-во объектов
        """
        return self.build_response(data={"count": self.filter_queryset(self.get_queryset()).count()})

    def partial_update(self, request, *args, **kwargs):
        """
        Запрет на частичное обновление
        """
        if not getattr(self, "allow_partial_update", True):
            raise MethodNotAllowed(self)
        return super().partial_update(request, *args, **kwargs)

    def get_client_ip(self):
        """
        Получение IP адреса клиента
        """

        return get_client_ip(request=self.request)

    def validate_captcha(self, token: str, ip: str = None) -> None:
        if not ip:
            ip = self.get_client_ip()

        validate_captcha(token=token, ip=ip)



from zs_utils.api.shopify.base_api import ShopifyAPI


class GetShopifyLocationListAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/location#get-locations
    """

    http_method = "GET"
    resource_method = "locations.json"


class GetShopifyLocationAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/location#get-locations-location-id
    """

    http_method = "GET"
    resource_method = "locations/{location_id}.json"
    required_params = ["location_id"]



from zs_utils.api.shopify.base_api import ShopifyAPI


class GetShopifyOrderListAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/order#get-orders?status=any
    """

    http_method = "GET"
    resource_method = "orders.json"
    allowed_params = [
        "attribution_app_id",
        "created_at_max",
        "created_at_min",
        "fields",
        "financial_status",
        "fulfillment_status",
        "ids",
        "limit",
        "processed_at_max",
        "processed_at_min",
        "since_id",
        "status",
        "updated_at_max",
        "updated_at_min",
        "page_info",
    ]


class GetShopifyOrderAPI(ShopifyAPI):
    """
    https://shopify.dev/docs/api/admin-rest/2023-10/resources/order#get-orders-order-id?fields=id,line-items,name,total-price
    """

    http_method = "GET"
    resource_method = "orders/{order_id}.json"
    required_params = ["order_id"]


class UpdateShopifyOrderAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/order#put-orders-order-id
    """

    http_method = "PUT"
    resource_method = "orders/{order_id}.json"
    payload_key = "order"
    required_params = ["order_id"]
    allowed_params = [
        "note",
        "note_attributes",
        "email",
        "phone",
        "buyer_accepts_marketing",
        "shipping_address",
        "customer",
        "tags",
        "metafields",
    ]


class CreateShopifyOrderAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/order#post-orders
    """

    http_method = "POST"
    resource_method = "orders.json"
    payload_key = "order"
    required_params = ["line_items"]
    allowed_params = [
        "email",
        "phone",
        "fulfillment_status",
        "inventory_behaviour",
        "send_receipt",
        "send_fulfillment_receipt",
        "fulfillments",
        "total_tax",
        "currency",
        "tax_lines",
        "customer",
        "financial_status",
        "billing_address",
        "shipping_address",
        "discount_codes",
        "note",
        "note_attributes",
        "buyer_accepts_marketing",
        "tags",
        "transactions",
        "metafields",
    ]


class DeleteShopifyOrderAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/order#delete-orders-order-id
    """

    http_method = "DELETE"
    resource_method = "orders/{order_id}.json"
    required_params = ["order_id"]


class GetShopifyOrdersCountAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/order#get-orders-count?status=any
    """

    http_method = "GET"
    resource_method = "orders/count.json"
    allowed_params = [
        "created_at_max",
        "created_at_min",
        "financial_status",
        "fulfillment_status",
        "status",
        "updated_at_max",
        "updated_at_min",
    ]



from model_utils.fields import MonitorField

from django.db import models
from django.contrib.auth.models import (
    AbstractBaseUser,
    BaseUserManager,
    PermissionsMixin,
)
from django.utils import timezone

from zs_utils.json_utils import CustomJSONEncoder


__all__ = [
    "AbstractZonesmartUser",
    "CustomUserManager",
]


class CustomUserManager(BaseUserManager):
    """
    Менеджер с возможностью создать пользователя
    """

    def create_user(
        self,
        email: str,
        password: str,
        is_staff: bool = False,
        is_superuser: bool = False,
        **kwargs,
    ):
        """
        Создать пользователя
        """
        user: AbstractZonesmartUser = self.model(email=email, is_staff=is_staff, is_superuser=is_superuser, **kwargs)
        user.set_password(password)
        user.save()
        return user


class AbstractZonesmartUser(AbstractBaseUser, PermissionsMixin):
    """
    Данные пользователя
    """

    USERNAME_FIELD = "email"
    EMAIL_FIELD = "email"
    REQUIRED_FIELDS = []

    email = models.EmailField(unique=True, verbose_name="E-mail")
    first_name = models.CharField(max_length=32, null=True, verbose_name="Имя")
    last_name = models.CharField(max_length=32, null=True, verbose_name="Фамилия")

    date_joined = models.DateTimeField(default=timezone.now, verbose_name="Дата регистрации")
    password_last_modified = MonitorField(monitor="password", verbose_name="Дата последнего изменения пароля")

    is_active = models.BooleanField(default=True, verbose_name="Активен")
    is_superuser = models.BooleanField(default=False, verbose_name="Супер юзер")
    is_staff = models.BooleanField(default=False, verbose_name="Администратор")

    extra_data = models.JSONField(
        default=dict,
        encoder=CustomJSONEncoder,
        verbose_name="Дополнительная информация",
    )

    objects = CustomUserManager()

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.email}"

    def set_password(self, raw_password: str, save: bool = False) -> None:
        super().set_password(raw_password=raw_password)

        self.password_last_modified = timezone.now()

        if save:
            self.save(update_fields=["password", "password_last_modified"])



from model_utils import Choices


ShippingOptionTypeEnum = Choices(
    ("DOMESTIC", "Внутренняя доставка"),
    ("INTERNATIONAL", "Международная доставка"),
)

ShippingCostTypeEnum = Choices(
    ("FLAT_RATE", "Flat rate"),
    ("CALCULATED", "Calculated"),
    ("NOT_SPECIFIED", "Not specified"),
)

CategoryTypeEnum = Choices(
    ("MOTORS_VEHICLES", "TRANSPORT", "Автотранспорт"),
    ("ALL_EXCLUDING_MOTORS_VEHICLES", "NOT_TRANSPORT", "Всё, кроме автотранспорта"),
)

RegionTypeEnum = Choices(
    ("COUNTRY", "Страна"),
    ("COUNTRY_REGION", "Страна или регион страны"),
    ("STATE_OR_PROVINCE", "Штат или процинция"),
    ("WORLD_REGION", "Регион мира"),
    ("WORLDWIDE", "Мир"),
)

PaymentInstrumentBrandEnum = Choices(
    ("AMERICAN_EXPRESS", "American Express"),
    ("DISCOVER", "Discover"),
    ("MASTERCARD", "MasterCard"),
    ("VISA", "Visa"),
)

FullPaymentDueInEnum = Choices(
    (3, "DAYS_3", "3 дня"),
    (7, "DAYS_7", "7 дней"),
    (10, "DAYS_10", "10 дней"),
    (14, "DAYS_14", "14 дней"),
)

ReturnMethodEnum = Choices(
    ("REPLACEMENT", "Замена"),
)

ReturnShippingCostPayerEnum = Choices(
    ("BUYER", "Покупатель"),
    ("SELLER", "Продавец"),
)

RecipientAccountReferenceTypeEnum = Choices(
    ("PAYPAL_EMAIL", "PayPal email"),
)

RefundMethodEnum = Choices(
    ("MERCHANDISE_CREDIT", "Продавец предлагает возврат кредитом"),
    ("MONEY_BACK", "Полный возврат продавцом"),
)

ReturnPeriodEnum = Choices(
    (14, "14 дней"),
    (30, "30 дней"),
    (60, "60 дней"),
)



"""
Docs:
https://developer.ebay.com/DevZone/guides/features-guide/default.html#development/feature-bestoffer.html
"""

from zs_utils.api.ebay.base_api import EbayTradingAPI
from .category import GetEbayCategoryFeatures
from .messages import AbstractSendMessage


__all__ = [
    "GetCategoryBestOfferFeatures",
    "GetListingBestOffers",
    "RespondToListingBestOffer",
    "SendBestOfferMessage",
]


class GetCategoryBestOfferFeatures(GetEbayCategoryFeatures):
    def get_params(self, category_id: int, **kwargs):
        kwargs["feature_ids"] = [
            "BestOfferEnabled",
            "BestOfferAutoDeclineEnabled",
            "BestOfferAutoAcceptEnabled",
        ]
        return super().get_params(category_id=category_id, **kwargs)


class GetListingBestOffers(EbayTradingAPI):
    """
    Docs:
    https://developer.ebay.com/DevZone/XML/docs/Reference/eBay/GetBestOffers.html
    """

    method_name = "GetBestOffers"

    def get_params(
        self,
        best_offer_id: int = None,
        item_id: int = None,
        active_only: bool = True,
        **kwargs,
    ):
        if active_only:
            best_offer_status = "Active"
        else:
            best_offer_status = "All"

        return {
            "BestOfferID": best_offer_id,
            "BestOfferStatus": best_offer_status,
            "ItemID": item_id,
            "DetailLevel": "ReturnAll",
        }


class RespondToListingBestOffer(EbayTradingAPI):
    """
    Docs:
    https://developer.ebay.com/devzone/xml/docs/reference/ebay/RespondToBestOffer.html
    """

    method_name = "RespondToBestOffer"

    def get_params(
        self,
        action: str,
        best_offer_id: int,
        item_id: int,
        counter_offer_price: float = None,
        counter_offer_quantity: int = None,
        seller_response: str = None,
        **kwargs,
    ):
        if not (action in ["Accept", "Counter", "Decline"]):
            raise AttributeError('Недопустимое значение параметра "action".')
        if best_offer_id and (not item_id):
            raise ValueError("Если задан 'best_offer_id', то должен быть задан 'item_id'.")

        return {
            "Action": action,
            "BestOfferID": best_offer_id,
            "ItemID": item_id,
            "CounterOfferPrice": counter_offer_price,
            "currencyID": "USD",
            "CounterOfferQuantity": counter_offer_quantity,
            "SellerResponse": seller_response,
        }


class SendBestOfferMessage(AbstractSendMessage):
    """
    AddMemberMessagesAAQToBidder:
    https://developer.ebay.com/Devzone/XML/docs/Reference/eBay/AddMemberMessagesAAQToBidder.html
    HINT: item needs to be tested
    """

    method_name = "AddMemberMessagesAAQToBidder"

    def get_params(self, **kwargs):
        params = super().get_params(**kwargs)
        params.update({"CorrelationID": "1"})
        return params



from django.utils import timezone
from django.db import transaction


from zs_utils.exceptions import CustomException
from zs_utils.base_support import constants, models, signals


__all__ = [
    "CommonSupportTicketService",
    "SupportTicketServiceException",
]


class SupportTicketServiceException(CustomException):
    pass


class CommonSupportTicketService:
    """
    Сервис для работы с тикетами (SupportTicket)
    """
    @classmethod
    def max_active_tickets_opened(cls, user) -> bool:
        """
        Проверка, что было открыто максимальное кол-во тикетов для пользователя user
        """
        return (
            constants.MAX_OPEN_TICKET
            <= models.SupportTicket.objects.filter(
                user=user, status__in=constants.SUPPORT_TICKET_ACTIVE_STATUSES_LIST
            ).count()
        )

    @classmethod
    def check_user_can_open_new_ticket(cls, user) -> None:
        """
        Проверка, что пользователь user может открыть новый тикет
        """
        if cls.max_active_tickets_opened(user=user):
            raise SupportTicketServiceException(message="Открыто максимальное количество заявок.")

    @classmethod
    def _generate_ticket_number(cls, user) -> str:
        """
        Генерация number для ticket
        """
        ticket_count = str(user.tickets.all().count() + 1)
        final_user_id = (4 - len(str(user.id))) * "0" + str(user.id)
        final_ticket_count = (3 - len(ticket_count)) * "0" + ticket_count
        return final_user_id + final_ticket_count

    @classmethod
    def create_ticket(
        cls,
        user,
        question_type: constants.SUPPORT_TICKET_QUESTION_TYPES,
        subject: str,
    ) -> models.SupportTicket:
        cls.check_user_can_open_new_ticket(user=user)

        return models.SupportTicket.objects.create(
            user=user,
            status=constants.SUPPORT_TICKET_STATUSES.PENDING,
            number=cls._generate_ticket_number(user=user),
            question_type=question_type,
            subject=subject,
        )

    @classmethod
    def create_ticket_message(
        cls,
        ticket: models.SupportTicket,
        sender,
        is_system: bool,
        text: str = None,
        files: list[models.SupportTicketMessageFile] = None,
    ) -> models.SupportTicketMessage:
        """
        Создание сообщения тикета и вложений
        """
        assert text or files, "Необходимо задать текст или файлы"
        if is_system:
            assert text and (not files), "Системное сообщение может быть только текстовым"
            assert not sender, "У системного сообщения не может быть отправителя"
        else:
            assert sender, "У пользовательского сообщения должен быть отправитель"

            if not ticket.is_active:
                raise SupportTicketServiceException(message="Заявка закрыта.")
            if sender not in [ticket.user, ticket.manager]:
                raise SupportTicketServiceException(message="Нельзя написать сообщение в чужую заявку.")

        signature = constants.SUPPORT_TICKET_MESSAGE_SIGNATURE
        if signature and text and (sender == ticket.manager):
            text += signature

        prev_ticket_client_status = ticket.client_status

        with transaction.atomic():
            ticket_message = models.SupportTicketMessage.objects.create(
                ticket=ticket,
                text=text,
                sender=sender,
                is_system=is_system,
                is_viewed=is_system,
            )
            if files:
                models.SupportTicketMessageFile.objects.filter(
                    id__in=[file.id for file in files]
                ).update(ticket_message=ticket_message)

        signals.support_ticket_message_created.send(
            sender=None, ticket_message=ticket_message, prev_ticket_client_status=prev_ticket_client_status
        )

        return ticket_message

    @classmethod
    def create_system_message_after_status_change(
        cls,
        ticket: models.SupportTicket,
        prev_status: constants.SUPPORT_TICKET_STATUSES,
    ) -> models.SupportTicketMessage:
        """
        Создание системного сообщения для тикета ticket после изменения его статуса
        """
        new_status = ticket.status
        text = None

        if prev_status in constants.SUPPORT_TICKET_ACTIVE_STATUSES:
            if new_status == constants.SUPPORT_TICKET_STATUSES.OPEN:
                text = constants.SUPPORT_TICKET_MESSAGES.OPEN
            elif new_status == constants.SUPPORT_TICKET_STATUSES.CLOSED_AUTO:
                text = constants.SUPPORT_TICKET_MESSAGES.CLOSED_AUTO
            elif new_status == constants.SUPPORT_TICKET_STATUSES.CLOSED_BY_USER:
                text = constants.SUPPORT_TICKET_MESSAGES.CLOSED_BY_USER
            elif new_status == constants.SUPPORT_TICKET_STATUSES.CLOSED_BY_MANAGER:
                text = constants.SUPPORT_TICKET_MESSAGES.CLOSED_BY_MANAGER
        else:
            if new_status == constants.SUPPORT_TICKET_STATUSES.OPEN:
                text = constants.SUPPORT_TICKET_MESSAGES.REOPEN

        if not text:
            raise NotImplementedError(f"Сообщение для смены статуса '{prev_status}' -> '{new_status}' неопределено.")

        return cls.create_ticket_message(ticket=ticket, text=text, sender=None, is_system=True)

    @classmethod
    def close_ticket(cls, user, ticket) -> None:
        if not ticket.is_active:
            raise SupportTicketServiceException(message="Заявка закрыта.")

        if ticket.user == user:
            status = constants.SUPPORT_TICKET_STATUSES.CLOSED_BY_USER
        elif ticket.manager == user:
            status = constants.SUPPORT_TICKET_STATUSES.CLOSED_BY_MANAGER
        else:
            raise SupportTicketServiceException(message="Нельзя закрыть чужую заявку.")

        cls.set_ticket_status(ticket=ticket, status=status)

    @classmethod
    def reopen_ticket(cls, user, ticket) -> None:
        cls.check_user_can_open_new_ticket(user=user)

        if ticket.is_active:
            raise SupportTicketServiceException(message="Заявка открыта.")

        cls.set_ticket_status(ticket=ticket, status=constants.SUPPORT_TICKET_STATUSES.OPEN)

    @classmethod
    def take_ticket(cls, manager, ticket) -> None:
        assert manager.is_staff

        if not ticket.is_active:
            raise SupportTicketServiceException(message="Заявка закрыта.")

        ticket.manager = manager
        ticket.save()

        cls.set_ticket_status(ticket=ticket, status=constants.SUPPORT_TICKET_STATUSES.OPEN)

    @classmethod
    def set_ticket_viewed(cls, user, ticket) -> None:
        ticket.messages.exclude(sender=user).update(is_viewed=True)

    @classmethod
    def set_ticket_status(cls, ticket, status: constants.SUPPORT_TICKET_STATUSES) -> None:
        """
        Установка нового статуса status для тикета ticket
        """

        if ticket.status != status:
            prev_status = ticket.status
            ticket.status = status
            ticket.save()

            signals.support_ticket_status_changed.send(sender=None, ticket=ticket, prev_status=prev_status)

    @classmethod
    def close_inactive_tickets(cls) -> None:
        """
        Закрытие всех неактивных обращений в поддержку после settings.AUTO_CLOSE_TICKET_AFTER
        """
        now = timezone.now()
        for ticket in models.SupportTicket.objects.filter(
                status=constants.SUPPORT_TICKET_ACTIVE_STATUSES.OPEN
        ):
            if ticket.manager:
                last_message = ticket.last_message
                # Последнее сообщение отправлено больше заданного кол-ва дней назад
                if (
                    last_message
                    and (last_message.sender == ticket.manager)
                    and (now > last_message.created + timezone.timedelta(days=constants.AUTO_CLOSE_TICKET_AFTER))
                ):
                    cls.set_ticket_status(ticket=ticket, status=constants.SUPPORT_TICKET_STATUSES.CLOSED_AUTO)



from zs_utils.api.apiship.base_api import ApishipAPI


class ApishipGetConnectionSchemasApi(ApishipAPI):
    resource_method = "connections/schemas"
    http_method = "GET"
    allowed_params = ["offset", "limit", "providerKey"]


class ApishipGetConnectionListApi(ApishipAPI):
    resource_method = "connections"
    http_method = "GET"
    allowed_params = ["offset", "limit", "filter"]


class ApishipCreateConnectionAPI(ApishipAPI):
    resource_method = "connections"
    http_method = "POST"
    required_params = [
        "providerKey",
        "name",
        "insuranceRate",
        "cashServiceRate",
        "connectParams",
        "isUseBaseConnect",
    ]


class ApishipGetConnectionAPI(ApishipAPI):
    resource_method = "connections/{connection_id}"
    http_method = "GET"
    required_params = ["connection_id"]


class ApishipUpdateConnectionAPI(ApishipAPI):
    resource_method = "connections/{connection_id}"
    http_method = "PUT"
    required_params = [
        "connection_id",
        "providerKey",
        "name",
        "insuranceRate",
        "cashServiceRate",
        "connectParams",
        "isUseBaseConnect",
    ]


class ApishipDeleteConnectionAPI(ApishipAPI):
    resource_method = "connections/{connection_id}"
    http_method = "DELETE"
    required_params = ["connection_id"]



from zs_utils.api.aliexpress.base_api import AliexpressAPI


class GetServiceTemplate(AliexpressAPI):
    resource_method = "aliexpress.postproduct.redefining.querypromisetemplatebyid"
    required_params = ["template_id"]



from zs_utils.api.ebay.base_api import EbayAPI


class GetUser(EbayAPI):
    """
    Docs:
    https://developer.ebay.com/api-docs/commerce/identity/resources/user/methods/getUser
    """

    production_api_url = "https://apiz.ebay.com/"
    sandbox_api_url = "https://apiz.sandbox.ebay.com/"
    resource_method = "commerce/identity/v1/user"
    http_method = "GET"



from zs_utils.api.shopify.base_api import ShopifyAPI


class CreateShopifyVariantAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/product-variant#post-products-product-id-variants
    """

    http_method = "POST"
    resource_method = "products/{product_id}/variants.json"
    payload_key = "variant"
    required_params = ["product_id", "option1"]
    allowed_params = [
        "price",
        "image_id",
        "metafields",
        "inventory_policy",
        "compare_at_price",
        "fulfillment_service",
        "taxable",
        "barcode",
        "image",
        "id",
        "option2",
        "option3",
        "weight",
        "weight_unit",
    ]


class DeleteShopifyVariantAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/product-variant#delete-products-product-id-variants-variant-id
    """

    http_method = "DELETE"
    resource_method = "products/{product_id}/variants/{variant_id}.json"
    required_params = ["product_id", "variant_id"]


class UpdateShopifyVariantAPI(ShopifyAPI):
    """
    https://shopify.dev/api/admin-rest/2022-01/resources/product-variant#put-variants-variant-id
    """

    http_method = "PUT"
    resource_method = "variants/{variant_id}.json"
    payload_key = "variant"
    required_params = ["variant_id"]
    allowed_params = [
        "option1",
        "price",
        "image_id",
        "metafields",
        "inventory_policy",
        "compare_at_price",
        "fulfillment_service",
        "taxable",
        "barcode",
        "image",
        "id",
        "option2",
        "option3",
        "weight",
        "weight_unit",
    ]



#!/usr/bin/env python
# coding=utf-8

import re
from decimal import Decimal as D
from functools import reduce

from zs_utils.unit_converter.data import PREFIXES, UNITS
from zs_utils.unit_converter.exceptions import UnitDoesntExistError
from zs_utils.unit_converter.units import Unit, Quantity


def parse(quantity: str) -> Quantity:
    return QuantityParser().parse(quantity)


class QuantityParser(object):
    quantity_re = re.compile(r"(?P<value>\d+[.,]?\d*)? *(?P<unit>.*)")

    def parse(self, quantity: str) -> Quantity:
        r = self.quantity_re.match(quantity)
        unit = UnitParser().parse(r.group("unit"))
        if r.group("value") is not None:
            if "," in r.group("value"):
                value = D(r.group("value").replace(",", "."))
            else:
                value = D(r.group("value"))
            return Quantity(value, unit)
        else:
            return unit


class UnitParser(object):
    unit_re = re.compile(r"(?P<unit>[a-zA-Z°Ωµ]+)\^?(?P<pow>[-+]?[0-9]*\.?[0-9]*)")

    def parse(self, unit: str) -> Unit:
        l_unit_s = self.unit_re.findall(unit)
        l_unit = [self._parse_unit(unit, power) for unit, power in l_unit_s]
        return reduce(lambda x, y: x * y, l_unit)

    def _parse_unit(self, unit: str, power: str) -> Unit:
        if power == "":
            return self._parse_simple_unit(unit)
        else:
            return self._parse_simple_unit(unit) ** float(power)

    @staticmethod
    def _parse_simple_unit(unit_s: str) -> Unit:
        """Parse a simple unit.

        In other word, parse an unit without a power value.
        """
        unit = None
        for prefix in PREFIXES.keys():
            if unit_s.startswith(prefix) and unit_s[len(prefix) :] in UNITS.keys():
                unit = UNITS[unit_s[len(prefix) :]]
                prefix = PREFIXES[prefix]
                break

        if unit is None:
            raise UnitDoesntExistError(unit_s)

        return prefix * unit



from .abstract import AbstractModel, XMLElement

from zs_utils.yandex_market_language.exceptions import ValidationError


class Parameter(AbstractModel):
    """
    Model of offer characteristics and parameters.

    Docs:
    https://yandex.ru/support/partnermarket/elements/param.html
    """

    __slots__ = ["name", "unit", "_value"]

    def __init__(self, name: str = "", value: str = None, unit: str = None):
        self.name = name
        self.value = value
        self.unit = unit

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, v):
        try:
            self._value = str(v)
        except (TypeError, ValueError):
            raise ValidationError("value must be a string")

    def create_dict(self, **kwargs) -> dict:
        return dict(name=self.name, value=self.value, unit=self.unit)

    def create_xml(self, **kwargs) -> XMLElement:
        attribs = {"name": self.name}
        if self.unit:
            attribs["unit"] = self.unit
        el = XMLElement("param", attribs)
        el.text = self.value
        return el

    @staticmethod
    def from_xml(el: XMLElement) -> "Parameter":
        return Parameter(value=el.text, **el.attrib)



from decimal import Decimal as D

from zs_utils.unit_converter.units import UnitPrefix, Unit

# ----------
# Prefix SI
# ----------
PREFIXES = {
    "y": UnitPrefix(symbol="y", name="yocto", factor=D("1E-24")),
    "z": UnitPrefix(symbol="z", name="zepto", factor=D("1E-21")),
    "a": UnitPrefix(symbol="a", name="atto", factor=D("1E-18")),
    "f": UnitPrefix(symbol="f", name="femto", factor=D("1E-15")),
    "p": UnitPrefix(symbol="p", name="pico", factor=D("1E-12")),
    "n": UnitPrefix(symbol="n", name="nano", factor=D("1E-9")),
    "µ": UnitPrefix(symbol="µ", name="micro", factor=D("1E-6")),
    "m": UnitPrefix(symbol="m", name="milli", factor=D("1E-3")),
    "c": UnitPrefix(symbol="c", name="centi", factor=D("1E-2")),
    "d": UnitPrefix(symbol="d", name="deci", factor=D("1E-1")),
    "": UnitPrefix(symbol="", name="", factor=D("1E0")),
    "da": UnitPrefix(symbol="da", name="deca", factor=D("1E+1")),
    "h": UnitPrefix(symbol="h", name="hecto", factor=D("1E+2")),
    "k": UnitPrefix(symbol="k", name="kilo", factor=D("1E+3")),
    "M": UnitPrefix(symbol="M", name="mega", factor=D("1E+6")),
    "G": UnitPrefix(symbol="G", name="giga", factor=D("1E+9")),
    "T": UnitPrefix(symbol="T", name="tera", factor=D("1E+12")),
    "P": UnitPrefix(symbol="P", name="peta", factor=D("1E+15")),
    "E": UnitPrefix(symbol="E", name="exa", factor=D("1E+18")),
    "Z": UnitPrefix(symbol="Z", name="zetta", factor=D("1E+21")),
    "Y": UnitPrefix(symbol="Y", name="yotta", factor=D("1E+24")),
}

# ----------
# Units
# ----------
UNITS = {
    # Basic SI units
    # --------------
    "m": Unit("m", "meter", L=1),
    "g": Unit("g", "gram", M=1, coef=D("1E-3")),
    "s": Unit("s", "second", T=1),
    "A": Unit("A", "ampere", I=1),
    "K": Unit("K", "kelvin", THETA=1),
    "mol": Unit("mol", "mole", N=1),
    "cd": Unit("cd", "candela", J=1),
    # Derived SI units
    # ----------------
    "Hz": Unit("Hz", "hertz", T=-1),
    "N": Unit("N", "newton", M=1, L=1, T=-2),
    "Pa": Unit("Pa", "pascal", M=1, L=-1, T=-2),
    "J": Unit("J", "joule", M=1, L=2, T=-2),
    "W": Unit("W", "watt", M=1, L=2, T=-3),
    "C": Unit("C", "coulomb", T=1, I=1),
    "V": Unit("V", "volt", M=1, L=2, T=-3, I=-1),
    "Ω": Unit("Ω", "ohm", M=1, L=2, T=-3, I=-2),
    "S": Unit("S", "siemens", M=-1, L=-2, T=3, I=2),
    "F": Unit("F", "farad", M=-1, L=-2, T=4, I=2),
    "T": Unit("T", "tesla", M=1, T=-2, I=-1),
    "Wb": Unit("Wb", "weber", M=1, L=2, T=-2, I=-1),
    "H": Unit("H", "henry", M=1, L=2, T=-2, I=-2),
    "°C": Unit("°C", "celsius", THETA=1, offset=D("273.15")),
    "rad": Unit("rad", "radian"),
    "sr": Unit("sr", "steradian"),
    "lm": Unit("lm", "lumen", J=1),
    "lx": Unit("lx", "lux", L=-2, J=1),
    "Bq": Unit("Bq", "becquerel", T=-1),
    "Gy": Unit("Gy", "gray", L=2, T=-2),
    "Sv": Unit("Sv", "sievert", L=2, T=-2),
    "kat": Unit("kat", "katal", T=-1, N=1),
    # Imperial system
    # ---------------
    "°F": Unit("°F", "fahrenheit", THETA=1, offset=D("273.15") - D("32") / D("1.8"), coef=D("1") / D("1.8")),
    "thou": Unit("th", "thou", L=1, coef=D("2.54E-5")),
    "inch": Unit("in", "inch", L=1, coef=D("2.54E-2")),
    "foot": Unit("ft", "foot", L=1, coef=D("3.048E-1")),
    "yard": Unit("yd", "yard", L=1, coef=D("9.144E-1")),
    "chain": Unit("ch", "chain", L=1, coef=D("20.1168")),
    "furlong": Unit("fur", "furlong", L=1, coef=D("201.168")),
    "mile": Unit("ml", "mile", L=1, coef=D("1609.344")),
    "league": Unit("lea", "league", L=1, coef=D("4828.032")),
    "oz": Unit("oz", "ounce", M=1, coef=D("2.835E-2")),
    "lb": Unit("lb", "pound", M=1, coef=D("4.536E-1")),
    # Miscellaneous units
    # -------------------
    "bar": Unit("bar", "bar", M=1, L=-1, T=-2, coef=D("1E5")),
    "min": Unit("min", "minute", T=1, coef=D("60")),
    "h": Unit("h", "hour", T=1, coef=D("3600")),
}



from django.apps import AppConfig


class SupportConfig(AppConfig):
    name = "zs_utils.base_support"
    label = "base_support"

    def ready(self) -> None:
        import zs_utils.base_support.signals  # noqa



import typing as t
from zs_utils.yandex_market_language import models
from zs_utils.yandex_market_language.models.abstract import XMLElement, XMLSubElement


class Promo(models.AbstractModel):
    """
    Docs: https://yandex.ru/support/partnermarket/elements/promo-gift.html
    """

    MAPPING = {
        "start-date": "start_date",
        "end-date": "end_date",
        "description": "description",
        "url": "url",
    }

    __slots__ = ["promo_id", "promo_type", "purchase", "promo_gifts", "start_date", "end_date", "description", "url"]

    def __init__(
        self,
        promo_id: str,
        promo_type: str,
        purchase: "Purchase",
        promo_gifts: t.List["PromoGift"],
        start_date=None,
        end_date=None,
        description=None,
        url=None,
    ):
        self.promo_id = promo_id
        self.promo_type = promo_type
        self.start_date = start_date
        self.end_date = end_date
        self.description = description
        self.url = url
        self.purchase = purchase
        self.promo_gifts = promo_gifts

    def create_dict(self, **kwargs) -> dict:
        return dict(
            promo_id=self.promo_id,
            promo_type=self.promo_type,
            start_date=self.start_date,
            end_date=self.end_date,
            description=self.description,
            url=self.url,
            purchase=self.purchase.to_dict(),
            promo_gifts=[pg.to_dict() for pg in self.promo_gifts],
        )

    def create_xml(self, **kwargs) -> XMLElement:
        attribs = {"id": self.promo_id, "type": self.promo_type}
        promo_el = XMLElement("promo", attribs)

        for tag, attr in self.MAPPING.items():
            v = getattr(self, attr)
            if v:
                el = XMLSubElement(promo_el, tag)
                el.text = v

        # Add purchase el
        self.purchase.to_xml(promo_el)

        # Add promo gifts
        promo_gifts_el = XMLSubElement(promo_el, "promo-gifts")
        for pg in self.promo_gifts:
            pg.to_xml(promo_gifts_el)

        return promo_el

    @classmethod
    def from_xml(cls, promo_el: XMLElement) -> "Promo":
        kwargs = dict(promo_id=promo_el.attrib.get("id"), promo_type=promo_el.attrib.get("type"), promo_gifts=[])

        for el in promo_el:
            if el.tag in cls.MAPPING:
                kwargs[cls.MAPPING[el.tag]] = el.text
            elif el.tag == "purchase":
                kwargs["purchase"] = Purchase.from_xml(el)
            elif el.tag == "promo-gifts":
                for pg_el in el:
                    kwargs["promo_gifts"].append(PromoGift.from_xml(pg_el))

        return Promo(**kwargs)


class Purchase(models.AbstractModel):
    """
    Docs: https://yandex.ru/support/partnermarket/elements/promo-gift.html
    """

    __slots__ = ["products", "required_quantity"]

    def __init__(self, products: t.List["Product"], required_quantity="1"):
        self.required_quantity = required_quantity
        self.products = products

    def create_dict(self, **kwargs) -> dict:
        return dict(required_quantity=self.required_quantity, products=[p.to_dict() for p in self.products])

    def create_xml(self, **kwargs) -> XMLElement:
        purchase_el = XMLElement("purchase")

        # Add required quantity el
        required_quantity_el = XMLSubElement(purchase_el, "required-quantity")
        required_quantity_el.text = self.required_quantity

        # Add products el
        for p in self.products:
            p.to_xml(purchase_el)

        return purchase_el

    @staticmethod
    def from_xml(purchase_el: XMLElement) -> "Purchase":
        kwargs = {"products": []}

        for el in purchase_el:
            if el.tag == "required-quantity":
                kwargs["required_quantity"] = el.text
            elif el.tag == "product":
                kwargs["products"].append(Product.from_xml(el))

        return Purchase(**kwargs)


class Product(models.AbstractModel):
    """
    Docs: https://yandex.ru/support/partnermarket/elements/promo-gift.html
    """

    __slots__ = ["offer_id", "category_id"]

    def __init__(self, offer_id: str = None, category_id: str = None):
        self.offer_id = offer_id
        self.category_id = category_id

    def create_dict(self, **kwargs) -> dict:
        return dict(
            offer_id=self.offer_id,
            category_id=self.category_id,
        )

    def create_xml(self, **kwargs) -> XMLElement:
        attribs = {}
        if self.offer_id:
            attribs["offer-id"] = self.offer_id
        if self.category_id:
            attribs["category-id"] = self.category_id
        return XMLElement("product", attribs)

    @staticmethod
    def from_xml(product_el: XMLElement) -> "Product":
        return Product(offer_id=product_el.attrib.get("offer-id"), category_id=product_el.attrib.get("category-id"))


class PromoGift(models.AbstractModel):
    """
    Docs:
    https://yandex.ru/support/partnermarket/elements/promo-gift.html
    """

    __slots__ = ["offer_id", "gift_id"]

    def __init__(self, offer_id: str = None, gift_id: str = None):
        self.offer_id = offer_id
        self.gift_id = gift_id

    def create_dict(self, **kwargs) -> dict:
        return dict(offer_id=self.offer_id, gift_id=self.gift_id)

    def create_xml(self, **kwargs) -> XMLElement:
        attribs = {}
        if self.offer_id:
            attribs["offer-id"] = self.offer_id
        elif self.gift_id:
            attribs["gift-id"] = self.gift_id
        return XMLElement("promo-gift", attribs)

    @staticmethod
    def from_xml(el: XMLElement) -> "PromoGift":
        return PromoGift(offer_id=el.attrib.get("offer-id"), gift_id=el.attrib.get("gift-id"))
